""" Job Schedulers are used to run job graphs.

Runs a job graph by calling simulate() on the python Model
objects in order.
"""
import itertools
import logging
import traceback
from collections import defaultdict

import networkx
from smif.data_layer import DataHandle
from smif.model import ModelOperation


class SerialJobScheduler(object):
    """Run JobGraphs produced by a :class:`~smif.controller.modelrun.ModelRun`
    """
    def __init__(self, store=None):
        self._status = defaultdict(lambda: 'unstarted')
        self._id_counter = itertools.count()
        self.logger = logging.getLogger(__name__)
        self.store = store

    def add(self, job_graph):
        """Add a JobGraph to the SerialJobScheduler and run directly

        Arguments
        ---------
        job_graph: :class:`networkx.graph`
        """
        job_graph_id = self._next_id()
        try:
            self._run(job_graph, job_graph_id)
        except Exception as ex:
            self._status[job_graph_id] = 'failed'
            traceback.print_exc()
            return job_graph_id, ex

        return job_graph_id, None

    def kill(self, job_graph_id):
        """Kill a job_graph that is already running - not implemented

        Parameters
        ----------
        job_graph_id: int
        """
        raise NotImplementedError

    def get_status(self, job_graph_id):
        """Get job graph status

        Parameters
        ----------
        job_graph_id: int

        Returns
        -------
        dict: A message containing the status

        Notes
        -----
        Possible statuses:

        unstarted:
            Job graph has not yet started
        running:
            Job graph is running
        done:
            Job graph was completed succesfully
        failed:
            Job graph completed running with an exit code
        """
        return {'status': self._status[job_graph_id]}

    def _run(self, job_graph, job_graph_id):
        """Run a job graph
        - sort the jobs into a single list
        - unpack model, data_handle and operation from each node
        """
        try:
            self.logger.profiling_start(
                'SerialJobScheduler._run()', 'graph_' + str(job_graph_id))
        except AttributeError:
            self.logger.info('START SerialJobScheduler._run():graph_' + str(job_graph_id))

        self._status[job_graph_id] = 'running'

        for job_node_id, job in self._get_run_order(job_graph):
            self._run_job(job_node_id, job)

        self._status[job_graph_id] = 'done'
        try:
            self.logger.profiling_stop(
                'SerialJobScheduler._run()', 'graph_' + str(job_graph_id))
        except AttributeError:
            self.logger.info(
                'STOP SerialJobScheduler._run()', 'graph_' + str(job_graph_id))

    def _run_job(self, job_node_id, job):
        self.logger.info("Job %s", job_node_id)
        try:
            self.logger.profiling_start('SerialJobScheduler._run()', 'job_' + job_node_id)
        except AttributeError:
            self.logger.info('START SerialJobScheduler._run()', 'job_' + job_node_id)

        model = job['model']
        data_handle = DataHandle(
            store=self.store,
            model=model,
            modelrun_name=job['modelrun_name'],
            current_timestep=job['current_timestep'],
            timesteps=job['timesteps'],
            decision_iteration=job['decision_iteration']
        )
        operation = job['operation']
        if operation is ModelOperation.BEFORE_MODEL_RUN:
            # before_model_run may not be implemented by all jobs
            if hasattr(model, "before_model_run"):
                model.before_model_run(data_handle)

        elif operation is ModelOperation.SIMULATE:
            model.simulate(data_handle)
        else:
            raise ValueError("Unrecognised operation: {}".format(operation))

        try:
            self.logger.profiling_stop('SerialJobScheduler._run()', 'job_' + job_node_id)
        except AttributeError:
            self.logger.info('STOP SerialJobScheduler._run()', 'job_' + job_node_id)

    def _next_id(self):
        return next(self._id_counter)

    @staticmethod
    def _get_run_order(graph):
        """Returns a list of jobs in a runnable order.

        Returns
        -------
        list
            A list of job nodes
        """
        try:
            # topological sort gives a single list from directed graph,
            # ignoring opportunities to run independent models in parallel
            run_order = networkx.topological_sort(graph)

            # list of Models (typically ScenarioModel and SectorModel)
            ordered_jobs = [(run, graph.nodes[run]) for run in run_order]
        except networkx.NetworkXUnfeasible:
            raise NotImplementedError("Job graphs must not contain cycles")

        return ordered_jobs



"""The decision module handles the three planning levels

Currently, only pre-specified planning is implemented.

The choices made in the three planning levels influence the set of interventions
and assets available within a model run.

The interventions available in a model run are stored in a dict keyed by name.

"""
__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"

import itertools
import os
from abc import ABCMeta, abstractmethod
from logging import getLogger
from types import MappingProxyType
from typing import Dict, List, Optional, Set, Tuple

from smif.data_layer.data_handle import ResultsHandle
from smif.data_layer.model_loader import ModelLoader
from smif.data_layer.store import Store
from smif.exception import SmifDataNotFoundError


class DecisionManager(object):
    """A DecisionManager is initialised with one or more model run strategies
    that refer to DecisionModules such as pre-specified planning,
    a rule-based models or multi-objective optimisation.
    These implementations influence the combination and ordering of decision
    iterations and model timesteps that need to be performed by the model runner.

    The DecisionManager presents a simple decision loop interface to the model runner,
    in the form of a generator which allows the model runner to iterate over the
    collection of independent simulations required at each step.

    The DecisionManager collates the output of the decision algorithms and
    writes the post-decision state through a ResultsHandle. This allows Models
    to access a given decision state (identified uniquely by timestep and
    decision iteration id).

    The :py:meth:`get_decisions` method passes a ResultsHandle down to a
    DecisionModule, allowing the DecisionModule to access model results from
    previous timesteps and decision iterations when making decisions

    Arguments
    ---------
    store: smif.data_layer.store.Store
    """

    def __init__(self, store: Store,
                 timesteps: List[int],
                 modelrun_name: str,
                 sos_model):

        self.logger = getLogger(__name__)

        self._store = store
        self._modelrun_name = modelrun_name
        self._sos_model = sos_model
        self._timesteps = timesteps
        self._decision_module = None

        self._register = {}  # type: Dict
        for sector_model in sos_model.sector_models:
            self._register.update(self._store.read_interventions(sector_model.name))
        self.planned_interventions = []  # type: List

        strategies = self._store.read_strategies(modelrun_name)
        self.logger.info("%s strategies found", len(strategies))
        self.pre_spec_planning = self._set_up_pre_spec_planning(modelrun_name, strategies)
        self._set_up_decision_modules(modelrun_name, strategies)

    def _set_up_pre_spec_planning(self,
                                  modelrun_name: str,
                                  strategies: List[Dict]):

        # Read in the historical interventions (initial conditions) directly
        initial_conditions = self._store.read_all_initial_conditions(modelrun_name)

        # Read in strategies
        planned_interventions = []  # type: List
        planned_interventions.extend(initial_conditions)

        for index, strategy in enumerate(strategies):
            # Extract pre-specified planning interventions
            if strategy['type'] == 'pre-specified-planning':

                msg = "Adding %s planned interventions to pre-specified-planning %s"
                self.logger.info(msg, len(strategy['interventions']), index)

                planned_interventions.extend(strategy['interventions'])

        # Create a Pre-Specified planning decision module with all
        # the planned interventions
        self.planned_interventions = self._tuplize_state(planned_interventions)

    def _set_up_decision_modules(self, modelrun_name, strategies):

        strategy_types = [x['type'] for x in strategies
                          if x['type'] != 'pre-specified-planning']
        if len(set(strategy_types)) > 1:
            msg = "Cannot use more the 2 type of strategy simultaneously"
            raise NotImplementedError(msg)

        for strategy in strategies:

            if strategy['type'] != 'pre-specified-planning':

                loader = ModelLoader()

                # absolute path to be crystal clear for ModelLoader when loading python class
                strategy['path'] = os.path.normpath(
                    os.path.join(self._store.model_base_folder, strategy['path']))
                strategy['timesteps'] = self._timesteps
                # Pass a reference to the register of interventions
                strategy['register'] = MappingProxyType(self.available_interventions)

                strategy['name'] = strategy['classname'] + '_' + strategy['type']

                self.logger.debug("Trying to load strategy: %s", strategy['name'])
                decision_module = loader.load(strategy)
                self._decision_module = decision_module  # type: DecisionModule

    @property
    def available_interventions(self) -> Dict[str, Dict]:
        """Returns a register of available interventions, i.e. those not planned
        """
        planned_names = set(name for build_year, name in self.planned_interventions)
        edited_register = {name: self._register[name]
                           for name in self._register.keys() -
                           planned_names}
        return edited_register

    def get_intervention(self, value):
        try:
            return self._register[value]
        except KeyError:
            msg = ""
            raise SmifDataNotFoundError(msg.format(value))

    def decision_loop(self):
        """Generate bundles of simulation steps to run

        Each call to this method returns a dict:

            {
                'decision_iterations': list of decision iterations (int),
                'timesteps': list of timesteps (int),
                'decision_links': (optional) dict of {
                    decision iteration in current bundle: decision iteration of previous bundle
                }
            }

        A bundle is composed differently according to the implementation of the
        contained DecisionModule.  For example:

        With only pre-specified planning, there is a single step in the loop,
        with a single decision iteration with timesteps covering the entire model horizon.

        With a rule based approach, there might be many steps in the loop, each with a single
        decision iteration and single timestep, moving on once some threshold is satisfied.

        With a genetic algorithm, there might be a configurable number of steps in the loop,
        each with multiple decision iterations (one for each member of the algorithm's
        population) and timesteps covering the entire model horizon.

        Implicitly, if the bundle returned in an iteration contains multiple decision
        iterations, they can be performed in parallel. If each decision iteration contains
        multiple timesteps, they can also be parallelised, so long as there are no temporal
        dependencies.

        Decision links are only required if the bundle timesteps do not start from the first
        timestep of the model horizon.
    """
        self.logger.debug("Calling decision loop")
        if self._decision_module:
            while True:
                bundle = next(self._decision_module)
                if bundle is None:
                    break
                self.logger.debug("Bundle returned: %s", bundle)
                self._get_and_save_bundle_decisions(bundle)
                yield bundle
        else:
            bundle = {
                'decision_iterations': [0],
                'timesteps': [x for x in self._timesteps]
            }
            self._get_and_save_bundle_decisions(bundle)
            yield bundle

    def _get_and_save_bundle_decisions(self, bundle):
        """Iterate over bundle and write decisions

        Arguments
        ---------
        bundle : dict

        Returns
        -------
        bundle : dict
            One definitive bundle across the decision modules

        """
        for iteration, timestep in itertools.product(
                bundle['decision_iterations'],
                bundle['timesteps']):
            self.get_and_save_decisions(iteration, timestep)

    def get_and_save_decisions(self, iteration, timestep):
        """Retrieves decisions for given timestep and decision iteration from each decision
        module and writes them to the store as state.

        Calls each contained DecisionModule for the given timestep and decision iteration in
        the `data_handle`, retrieving a list of decision dicts (keyed by intervention name and
        build year).

        These decisions are then written to a state file using the data store.

        Arguments
        ---------
        timestep : int
        iteration : int

        Notes
        -----
        State contains all intervention names which are present in the system at
        the given ``timestep`` for the current ``iteration``. This must include
        planned interventions from a previous timestep that are still within
        their lifetime, and interventions picked by a decision module in the
        previous timesteps.

        After loading all historical interventions, and screening them to remove
        interventions from the previous timestep that have reached the end
        of their lifetime, new decisions are added to the list of current
        interventions.

        Finally, the new state is written to the store.
        """
        results_handle = ResultsHandle(
            store=self._store,
            modelrun_name=self._modelrun_name,
            sos_model=self._sos_model,
            current_timestep=timestep,
            timesteps=self._timesteps,
            decision_iteration=iteration
        )

        # Decision module overrides pre-specified planning for obtaining state
        # from previous iteration
        pre_decision_state = set()
        if self._decision_module:
            previous_state = self._get_previous_state(self._decision_module, results_handle)
            pre_decision_state.update(previous_state)

        msg = "Pre-decision state at timestep %s and iteration %s:\n%s"
        self.logger.debug(msg,
                          timestep, iteration, pre_decision_state)

        new_decisions = set()
        if self._decision_module:
            decisions = self._get_decisions(self._decision_module,
                                            results_handle)
            new_decisions.update(decisions)

        self.logger.debug("New decisions at timestep %s and iteration %s:\n%s",
                          timestep, iteration, new_decisions)

        # Post decision state is the union of the pre decision state set, the
        # new decision set and the set of planned interventions
        post_decision_state = (pre_decision_state
                               | new_decisions
                               | set(self.planned_interventions))
        post_decision_state = self.retire_interventions(post_decision_state, timestep)
        post_decision_state = self._untuplize_state(post_decision_state)

        self.logger.debug("Post-decision state at timestep %s and iteration %s:\n%s",
                          timestep, iteration, post_decision_state)

        self._store.write_state(post_decision_state, self._modelrun_name, timestep, iteration)

    def retire_interventions(self, state: List[Tuple[int, str]],
                             timestep: int) -> List[Tuple[int, str]]:

        alive = []
        for intervention in state:
            build_year = int(intervention[0])
            data = self._register[intervention[1]]
            lifetime = data['technical_lifetime']['value']
            if (self.buildable(build_year, timestep) and
                    self.within_lifetime(build_year, timestep, lifetime)):
                alive.append(intervention)
        return alive

    def _get_decisions(self,
                       decision_module: 'DecisionModule',
                       results_handle: ResultsHandle) -> List[Tuple[int, str]]:
        decisions = decision_module.get_decision(results_handle)
        return self._tuplize_state(decisions)

    def _get_previous_state(self,
                            decision_module: 'DecisionModule',
                            results_handle: ResultsHandle) -> List[Tuple[int, str]]:
        state_dict = decision_module.get_previous_state(results_handle)
        return self._tuplize_state(state_dict)

    @staticmethod
    def _tuplize_state(state: List[Dict]) -> List[Tuple[int, str]]:
        return [(x['build_year'], x['name']) for x in state]

    @staticmethod
    def _untuplize_state(state: List[Tuple[int, str]]) -> List[Dict]:
        return [{'build_year': x[0], 'name': x[1]} for x in state]

    def buildable(self, build_year, timestep) -> bool:
        """Interventions are deemed available if build_year is less than next timestep

        For example, if `a` is built in 2011 and timesteps are
        [2005, 2010, 2015, 2020] then buildable returns True for timesteps
        2010, 2015 and 2020 and False for 2005.

        Arguments
        ---------
        build_year: int
            The build year of the intervention
        timestep: int
            The current timestep
        """
        if not isinstance(build_year, (int, float)):
            msg = "Build Year should be an integer but is a {}"
            raise TypeError(msg.format(type(build_year)))
        if timestep not in self._timesteps:
            raise ValueError("Timestep not in model timesteps")
        index = self._timesteps.index(timestep)
        if index == len(self._timesteps) - 1:
            next_year = timestep + 1
        else:
            next_year = self._timesteps[index + 1]

        if int(build_year) < next_year:
            return True
        else:
            return False

    @staticmethod
    def within_lifetime(build_year, timestep, lifetime) -> bool:
        """Interventions are deemed active if build_year + lifetime >= timestep

        Arguments
        ---------
        build_year : int
        timestep : int
        lifetime : int

        Returns
        -------
        bool
        """
        if not isinstance(build_year, (int, float)):
            msg = "Build Year should be an integer but is a {}"
            raise TypeError(msg.format(type(build_year)))

        try:
            build_year = int(build_year)
        except ValueError:
            raise ValueError(
                "A build year must be a valid integer. Received {}.".format(build_year))

        try:
            lifetime = int(lifetime)
        except ValueError:
            lifetime = float("inf")
        if lifetime < 0:
            msg = "The value of lifetime cannot be negative"
            raise ValueError(msg)
        if timestep <= build_year + lifetime:
            return True
        else:
            return False


class DecisionModule(metaclass=ABCMeta):
    """Abstract class which provides the interface to user defined decision modules.

    These mechanisms could include a Rule-based Approach or Multi-objective Optimisation.

    This class provides two main methods, ``__next__`` which is normally
    called implicitly as a call to the class as an iterator, and ``get_decision()``
    which takes as arguments a smif.data_layer.data_handle.ResultsHandle object.

    Arguments
    ---------
    timesteps : list
        A list of planning timesteps
    register : dict
        Reference to a dict of iterventions
    """

    """Current iteration of the decision module
    """
    def __init__(self, timesteps: List[int], register: MappingProxyType):
        self.timesteps = timesteps
        self._register = register
        self.logger = getLogger(__name__)
        self._decisions = set()  # type: Set

    def __next__(self) -> List[Dict]:
        return self._get_next_decision_iteration()

    @property
    def first_timestep(self) -> int:
        return min(self.timesteps)

    @property
    def last_timestep(self) -> int:
        return max(self.timesteps)

    @abstractmethod
    def get_previous_state(self, results_handle: ResultsHandle) -> List[Dict]:
        """Return the state of the previous timestep
        """
        raise NotImplementedError

    def available_interventions(self, state: List[Dict]) -> List:
        """Return the collection of available interventions

        Available interventions are the subset of interventions that have not
        been implemented in a prior iteration or timestep

        Returns
        -------
        List
        """
        return [name for name in self._register.keys()
                - set([x['name'] for x in state])]

    def get_intervention(self, name):
        """Return an intervention dict

        Returns
        -------
        dict
        """
        try:
            return self._register[name]
        except KeyError:
            msg = "Intervention '{}' is not found in the list of available interventions"
            raise SmifDataNotFoundError(msg.format(name))

    @abstractmethod
    def _get_next_decision_iteration(self) -> List[Dict]:
        """Implement to return the next decision iteration

        Within a list of decision-iteration/timestep pairs, the assumption is
        that all decision iterations can be run in parallel
        (otherwise only one will be returned) and within a decision interation,
        all timesteps may be run in parallel as long as there are no
        inter-timestep state dependencies

        Returns
        -------
        dict
            Yields a dictionary keyed by ``decision iterations`` whose values
            contain a list of iteration integers, ``timesteps`` whose values
            contain a list of timesteps run in each decision iteration and the
            optional ``decision_links`` which link the decision interation of the
            current bundle to that of the previous bundle::

            {
                'decision_iterations': list of decision iterations (int),
                'timesteps': list of timesteps (int),
                'decision_links': (optional) dict of {
                    decision iteration in current bundle: decision iteration of previous bundle
                }
            }
        """
        raise NotImplementedError

    @abstractmethod
    def get_decision(self, results_handle: ResultsHandle) -> List[Dict]:
        """Return decisions for a given timestep and decision iteration

        Parameters
        ----------
        results_handle : smif.data_layer.data_handle.ResultsHandle

        Returns
        -------
        list of dict

        Examples
        --------
        >>> register = {'intervention_a': {'capital_cost': {'value': 1234}}}
        >>> dm = DecisionModule([2010, 2015], register)
        >>> dm.get_decision(results_handle)
        [{'name': 'intervention_a', 'build_year': 2010}])
        """
        raise NotImplementedError


class RuleBased(DecisionModule):
    """Rule-base decision modules
    """

    def __init__(self, timesteps, register):
        super().__init__(timesteps, register)
        self.satisfied = False  # type: bool
        self.current_timestep = self.first_timestep  # type: int
        self.current_iteration = 0  # type: int
        # keep internal account of max iteration reached per timestep
        self._max_iteration_by_timestep = {self.first_timestep: 0}
        self.logger = getLogger(__name__)

    def get_previous_iteration_timestep(self) -> Optional[Tuple[int, int]]:
        """Returns the timestep, iteration pair that describes the previous
        iteration

        Returns
        -------
        tuple
            Contains (timestep, iteration)
        """
        if self.current_iteration > 1:
            iteration = self.current_iteration - 1

            if self.current_timestep == self.first_timestep:
                timestep = self.current_timestep
            elif (iteration == self._max_iteration_by_timestep[self.previous_timestep]):
                timestep = self.previous_timestep
            elif (iteration >= self._max_iteration_by_timestep[self.previous_timestep]):
                timestep = self.current_timestep
        else:
            return None
        return timestep, iteration

    def get_previous_state(self, results_handle: ResultsHandle) -> List[Dict]:

        timestep_iteration = self.get_previous_iteration_timestep()
        if timestep_iteration:
            timestep, iteration = timestep_iteration
            return results_handle.get_state(timestep, iteration)
        else:
            return []

    @property
    def next_timestep(self):
        index_current_timestep = self.timesteps.index(self.current_timestep)
        try:
            return self.timesteps[index_current_timestep + 1]
        except IndexError:
            return None

    @property
    def previous_timestep(self):
        index_current_timestep = self.timesteps.index(self.current_timestep)
        if index_current_timestep > 0:
            return self.timesteps[index_current_timestep - 1]
        else:
            return None

    def _get_next_decision_iteration(self):
        if self.satisfied and (self.current_timestep == self.last_timestep):
            return None
        elif self.satisfied and (self.current_timestep < self.last_timestep):
            self._max_iteration_by_timestep[self.current_timestep] = \
                self.current_iteration
            self.satisfied = False
            self.current_timestep = self.next_timestep
            self.current_iteration += 1
            return self._make_bundle()
        else:
            self.current_iteration += 1
            return self._make_bundle()

    def get_previous_year_iteration(self):
        iteration = self._max_iteration_by_timestep[self.previous_timestep]
        return iteration

    def _make_bundle(self):
        bundle = {'decision_iterations': [self.current_iteration],
                  'timesteps': [self.current_timestep]}

        if self.current_timestep > self.first_timestep:
            bundle['decision_links'] = {
                self.current_iteration: self._max_iteration_by_timestep[
                    self.previous_timestep]
            }
        return bundle

    def get_decision(self, results_handle) -> List[Dict]:
        return []



"""Test the Results interface
"""

import os
from collections import OrderedDict

import numpy as np
import pandas as pd
from pytest import fixture, raises
from smif.data_layer import DataArray, Results
from smif.exception import SmifDataNotFoundError
from smif.metadata import Spec


@fixture
def results_no_results(empty_store):
    """Results fixture with a model run and fictional results
    """
    empty_store.write_dimension({
        'name': 'sample_dim',
        'elements': [{'name': 'a'}, {'name': 'b'}]
    })
    empty_store.write_dimension({
        'name': 'sample_dim_colour',
        'elements': [{'name': 'red'}, {'name': 'green'}, {'name': 'blue'}]
    })
    sample_output = {
        'name': 'sample_output',
        'dtype': 'int',
        'dims': ['sample_dim', 'sample_dim_colour'],
        'coords': {
            'sample_dim': [{'name': 'a'}, {'name': 'b'}],
            'sample_dim_colour': [{'name': 'red'}, {'name': 'green'}, {'name': 'blue'}],
        },
        'unit': 'm'
    }
    scenarios_1 = {
        'a_scenario': 'a_variant_1',
        'b_scenario': 'b_variant_1',
    }
    scenarios_2 = {
        'a_scenario': 'a_variant_2',
        'b_scenario': 'b_variant_2',
    }
    empty_store.write_model({
        'name': 'a_model',
        'description': "Sample model",
        'classname': 'DoesNotExist',
        'path': '/dev/null',
        'inputs': [],
        'outputs': [sample_output],
        'parameters': [],
        'interventions': [],
        'initial_conditions': []
    })
    empty_store.write_model({
        'name': 'b_model',
        'description': "Second sample model",
        'classname': 'DoesNotExist',
        'path': '/dev/null',
        'inputs': [],
        'outputs': [sample_output],
        'parameters': [],
        'interventions': [],
        'initial_conditions': []
    })
    empty_store.write_sos_model({
        'name': 'a_sos_model',
        'description': 'Sample SoS',
        'sector_models': ['a_model', 'b_model'],
        'scenarios': [],
        'scenario_dependencies': [],
        'model_dependencies': [],
        'narratives': []
    })
    empty_store.write_model_run({
        'name': 'model_run_1',
        'description': 'Sample model run',
        'timesteps': [2010, 2015, 2020, 2025, 2030],
        'sos_model': 'a_sos_model',
        'scenarios': scenarios_1,
        'strategies': [],
        'narratives': {}
    })
    empty_store.write_model_run({
        'name': 'model_run_2',
        'description': 'Sample model run',
        'timesteps': [2010, 2015, 2020, 2025, 2030],
        'sos_model': 'a_sos_model',
        'scenarios': scenarios_2,
        'strategies': [],
        'narratives': {}
    })

    return Results(store=empty_store)


@fixture
def results_with_results(results_no_results):
    sample_output = {
        'name': 'sample_output',
        'dtype': 'int',
        'dims': ['sample_dim', 'sample_dim_colour'],
        'coords': {
            'sample_dim': [{'name': 'a'}, {'name': 'b'}],
            'sample_dim_colour': [{'name': 'red'}, {'name': 'green'}, {'name': 'blue'}],
        },
        'unit': 'm'
    }

    spec = Spec.from_dict(sample_output)
    data = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.int32)
    sample_results = DataArray(spec, data)

    results_no_results._store.write_results(sample_results, 'model_run_1', 'a_model', 2010, 0)
    results_no_results._store.write_results(sample_results, 'model_run_1', 'a_model', 2015, 0)
    results_no_results._store.write_results(sample_results, 'model_run_1', 'a_model', 2020, 0)
    results_no_results._store.write_results(sample_results, 'model_run_1', 'a_model', 2015, 1)
    results_no_results._store.write_results(sample_results, 'model_run_1', 'a_model', 2020, 1)
    results_no_results._store.write_results(sample_results, 'model_run_1', 'a_model', 2015, 2)
    results_no_results._store.write_results(sample_results, 'model_run_1', 'a_model', 2020, 2)

    results_no_results._store.write_results(sample_results, 'model_run_1', 'b_model', 2010, 0)
    results_no_results._store.write_results(sample_results, 'model_run_1', 'b_model', 2015, 0)
    results_no_results._store.write_results(sample_results, 'model_run_1', 'b_model', 2020, 0)
    results_no_results._store.write_results(sample_results, 'model_run_1', 'b_model', 2025, 0)
    results_no_results._store.write_results(sample_results, 'model_run_1', 'b_model', 2030, 0)

    results_no_results._store.write_results(sample_results, 'model_run_2', 'b_model', 2010, 0)
    results_no_results._store.write_results(sample_results, 'model_run_2', 'b_model', 2015, 0)
    results_no_results._store.write_results(sample_results, 'model_run_2', 'b_model', 2020, 0)
    results_no_results._store.write_results(sample_results, 'model_run_2', 'b_model', 2025, 0)
    results_no_results._store.write_results(sample_results, 'model_run_2', 'b_model', 2030, 0)

    return results_no_results


class TestNoResults:

    def test_exceptions(self, empty_store):
        # No arguments is not allowed
        with raises(TypeError) as ex:
            Results()
        assert "missing 1 required positional argument: 'store'" in str(ex.value)

        # Check that constructing with just a store works fine
        Results(store=empty_store)

        # Check that valid configurations do work (but expect a SmifDataNotFoundError
        # because the store creation will fall over
        with raises(SmifDataNotFoundError) as ex:
            Results(store={'interface': 'local_csv', 'dir': '.'})
        assert 'Expected data folder' in str(ex.value)

        with raises(SmifDataNotFoundError) as ex:
            Results(store={'interface': 'local_parquet', 'dir': '.'})
        assert 'Expected data folder' in str(ex.value)

        # Interface left blank will default to local_csv
        with raises(SmifDataNotFoundError) as ex:
            Results(store={'dir': '.'})
        assert 'Expected data folder' in str(ex.value)

        # Dir left blank will default to '.'
        with raises(SmifDataNotFoundError) as ex:
            Results(store={'interface': 'local_parquet'})
        assert 'Expected data folder' in str(ex.value)

        # Invalid interface will raise a ValueError
        with raises(ValueError) as ex:
            Results(store={'interface': 'invalid', 'dir': '.'})
        assert 'Unsupported interface "invalid"' in str(ex.value)

        # Invalid directory will raise a ValueError
        with raises(ValueError) as ex:
            invalid_dir = os.path.join(os.path.dirname(__file__), 'does', 'not', 'exist')
            Results(store={'interface': 'local_csv', 'dir': invalid_dir})
        assert 'to be a valid directory' in str(ex.value)

    def test_list_model_runs(self, results_no_results):
        assert results_no_results.list_model_runs() == ['model_run_1', 'model_run_2']

    def test_list_no_model_runs(self, empty_store):
        # Should be no model runs in an empty Results()
        results = Results(store=empty_store)
        assert results.list_model_runs() == []

    def test_list_outputs(self, results_no_results):
        assert results_no_results.list_outputs('a_model') == ['sample_output']

    def test_list_sector_models(self, results_no_results):
        assert results_no_results.list_sector_models('model_run_1') == ['a_model', 'b_model']
        assert results_no_results.list_sector_models('model_run_2') == ['a_model', 'b_model']

    def test_list_scenarios(self, results_no_results):
        scenarios_dict = results_no_results.list_scenarios('model_run_1')
        assert scenarios_dict['a_scenario'] == 'a_variant_1'
        assert scenarios_dict['b_scenario'] == 'b_variant_1'

        scenarios_dict = results_no_results.list_scenarios('model_run_2')
        assert scenarios_dict['a_scenario'] == 'a_variant_2'
        assert scenarios_dict['b_scenario'] == 'b_variant_2'

    def test_list_scenario_outputs(self, results_no_results):
        store = results_no_results._store
        store.write_scenario({
            'name': 'a_scenario',
            'provides': [{'name': 'a_provides'}, {'name': 'b_provides'}]
        })

        assert results_no_results.list_scenario_outputs('a_scenario') == ['a_provides',
                                                                          'b_provides']

    def test_available_results(self, results_no_results):
        available = results_no_results.available_results('model_run_1')

        assert available['model_run'] == 'model_run_1'
        assert available['sos_model'] == 'a_sos_model'
        assert available['sector_models'] == {}
        assert available['scenarios']['a_scenario'] == 'a_variant_1'
        assert available['scenarios']['b_scenario'] == 'b_variant_1'


class TestSomeResults:

    def test_available_results(self, results_with_results):
        available = results_with_results.available_results('model_run_1')

        assert available['model_run'] == 'model_run_1'
        assert available['sos_model'] == 'a_sos_model'
        assert available['scenarios']['a_scenario'] == 'a_variant_1'
        assert available['scenarios']['b_scenario'] == 'b_variant_1'

        sec_models = available['sector_models']
        assert sorted(sec_models.keys()) == ['a_model', 'b_model']

        # Check a_model outputs are correct
        outputs_a = sec_models['a_model']['outputs']
        assert sorted(outputs_a.keys()) == ['sample_output']

        output_answer_a = {0: [2010, 2015, 2020], 1: [2015, 2020], 2: [2015, 2020]}
        assert outputs_a['sample_output'] == output_answer_a

        # Check b_model outputs are correct
        outputs_b = sec_models['b_model']['outputs']
        assert sorted(outputs_b.keys()) == ['sample_output']

        output_answer_b = {0: [2010, 2015, 2020, 2025, 2030]}
        assert outputs_b['sample_output'] == output_answer_b

        available = results_with_results.available_results('model_run_2')

        assert available['model_run'] == 'model_run_2'
        assert available['sos_model'] == 'a_sos_model'

        sec_models = available['sector_models']
        assert sorted(sec_models.keys()) == ['b_model']

        # Check a_model outputs are correct
        outputs = sec_models['b_model']['outputs']
        assert sorted(outputs_a.keys()) == ['sample_output']

        output_answer = {0: [2010, 2015, 2020, 2025, 2030]}
        assert outputs['sample_output'] == output_answer

    def test_read_validate_names(self, results_with_results):
        # Passing anything other than one sector model or output is current not implemented
        with raises(NotImplementedError) as e:
            results_with_results.read_results(
                model_run_names=['model_run_1', 'model_run_2'],
                model_names=[],
                output_names=['sample_output']
            )
        assert 'requires exactly one sector model' in str(e.value)

        with raises(NotImplementedError) as e:
            results_with_results.read_results(
                model_run_names=['model_run_1', 'model_run_2'],
                model_names=['a_model', 'b_model'],
                output_names=['one']
            )
        assert 'requires exactly one sector model' in str(e.value)

        with raises(ValueError) as e:
            results_with_results.read_results(
                model_run_names=[],
                model_names=['a_model'],
                output_names=['sample_output']
            )
        assert 'requires at least one sector model name' in str(e.value)

        with raises(ValueError) as e:
            results_with_results.read_results(
                model_run_names=['model_run_1'],
                model_names=['a_model'],
                output_names=[]
            )
        assert 'requires at least one output name' in str(e.value)

    def test_read(self, results_with_results):
        # Read one model run and one output
        results_data = results_with_results.read_results(
            model_run_names=['model_run_1'],
            model_names=['a_model'],
            output_names=['sample_output']
        )

        expected = pd.DataFrame(
            OrderedDict([
                ('model_run', 'model_run_1'),
                ('timestep', [2010] * 6 + [2015] * 18 + [2020] * 18),
                ('decision', [0] * 12 + [1] * 6 + [2] * 6 + [0] * 6 + [1] * 6 + [2] * 6),
                ('sample_dim', ['a', 'a', 'a', 'b', 'b', 'b'] * 7),
                ('sample_dim_colour', ['red', 'green', 'blue'] * 14),
                ('sample_output', np.asarray([1, 2, 3, 4, 5, 6] * 7, dtype=np.int32)),
            ])
        )

        pd.testing.assert_frame_equal(results_data, expected)

        # Read two model runs and one output
        results_data = results_with_results.read_results(
            model_run_names=['model_run_1', 'model_run_2'],
            model_names=['b_model'],
            output_names=['sample_output']
        )

        expected = pd.DataFrame(
            OrderedDict([
                ('model_run', ['model_run_1'] * 30 + ['model_run_2'] * 30),
                ('timestep', [2010] * 6 + [2015] * 6 + [2020] * 6 + [2025] * 6 + [2030] * 6 +
                             [2010] * 6 + [2015] * 6 + [2020] * 6 + [2025] * 6 + [2030] * 6),
                ('decision', 0),
                ('sample_dim', ['a', 'a', 'a', 'b', 'b', 'b'] * 10),
                ('sample_dim_colour', ['red', 'green', 'blue'] * 20),
                ('sample_output', np.asarray([1, 2, 3, 4, 5, 6] * 10, dtype=np.int32)),
            ])
        )

        pd.testing.assert_frame_equal(results_data, expected)


class TestReadScenarios:

    def test_read_scenario_variant_data(self, results_no_results, model_run, sample_dimensions,
                                        scenario,
                                        sample_scenario_data):
        store = results_no_results._store

        # Setup
        for dim in sample_dimensions:
            store.write_dimension(dim)
        store.write_scenario(scenario)
        # Pick out single sample
        key = next(iter(sample_scenario_data))
        scenario_name, variant_name, variable = key
        scenario_variant_data = sample_scenario_data[key]
        # Write
        store.write_scenario_variant_data(scenario_name, variant_name, scenario_variant_data)
        # End setup

        scenario_data_frame = results_no_results.read_scenario_data(
            scenario_name, variant_name, variable, [2015, 2016]
        )

        expected = pd.DataFrame(
            OrderedDict([
                ('timestep', [2015, 2015, 2016, 2016]),
                ('lad', ['a', 'b', 'a', 'b']),
                ('mortality', scenario_variant_data.data.flatten()),
            ])
        )

        pd.testing.assert_frame_equal(scenario_data_frame, expected)



"""A data store holds the bulk of model setup, intermediate and output data:
- scenario variant and narrative data (including parameter defaults)
- model interventions, initial conditions and state
- conversion coefficients
- results
"""
from abc import ABCMeta, abstractmethod
from typing import Dict, List

from smif.data_layer.data_array import DataArray
from smif.exception import SmifDataMismatchError, SmifDataNotFoundError
from smif.metadata import Spec


class DataStore(metaclass=ABCMeta):
    """A DataStore must implement each of the abstract methods defined in this interface
    """
    # region DataArray
    @abstractmethod
    def read_scenario_variant_data(
            self, key, spec, timestep=None, timesteps=None) -> DataArray:
        """Read scenario variant data array.

        If a single timestep is specified, the spec MAY include 'timestep' as a dimension,
        which should match the timestep specified.

        If multiple timesteps are specified, the spec MUST include 'timestep' as a dimension,
        which should match the timesteps specified.

        If timestep and timesteps are None, read all available timesteps. Whether or not the
        spec includes 'timestep' as a dimension, the returned DataArray will include a
        'timestep' dimension with all available timesteps included.

        Parameters
        ----------
        key : str
        spec : ~smif.metadata.spec.Spec
        timestep : int (optional)
            If set, read data for single timestep
        timesteps : list[int] (optional)
            If set, read data for specified timesteps

        Returns
        -------
        data_array : ~smif.data_layer.data_array.DataArray
        """

    @abstractmethod
    def scenario_variant_data_exists(self, key) -> bool:
        """Test if scenario variant data exists

        Parameters
        ----------
        key : str

        Returns
        -------
        bool
        """

    @abstractmethod
    def write_scenario_variant_data(self, key, data_array):
        """Write data array

        Parameters
        ----------
        key : str
        data_array : ~smif.data_layer.data_array.DataArray
        """

    @abstractmethod
    def read_narrative_variant_data(self, key, spec, timestep=None):
        """Read data array

        Parameters
        ----------
        key : str
        spec : ~smif.metadata.spec.Spec
        timestep : int (optional)
            If None, read data for all timesteps

        Returns
        -------
        data_array : ~smif.data_layer.data_array.DataArray
        """

    @abstractmethod
    def write_narrative_variant_data(self, key, data_array):
        """Write data array

        Parameters
        ----------
        key : str
        data_array : ~smif.data_layer.data_array.DataArray
        """

    @abstractmethod
    def read_model_parameter_default(self, key, spec):
        """Read data array

        Parameters
        ----------
        key : str
        spec : ~smif.metadata.spec.Spec

        Returns
        -------
        data_array : ~smif.data_layer.data_array.DataArray
        """

    @abstractmethod
    def write_model_parameter_default(self, key, data_array):
        """Read data array

        Parameters
        ----------
        key : str
        data_array : ~smif.data_layer.data_array.DataArray

        Returns
        -------
        data_array : ~smif.data_layer.data_array.DataArray
        """
    # endregion

    # region Interventions
    @abstractmethod
    def read_interventions(self, key):
        """Read interventions data for `key`

        Parameters
        ----------
        key : str

        Returns
        -------
        dict[str, dict]
            A dict of intervention dictionaries containing intervention
            attributes keyed by intervention name
        """

    @abstractmethod
    def write_interventions(self, key, interventions):
        """Write interventions data for `key`

        Parameters
        ----------
        key : str
        interventions : dict[str, dict]
        """

    @abstractmethod
    def read_initial_conditions(self, key) -> List[Dict]:
        """Read historical interventions for `key`

        Parameters
        ----------
        key : str

        Returns
        -------
        list[dict]
        """

    @abstractmethod
    def write_initial_conditions(self, key, initial_conditions):
        """Write historical interventions for `key`

        Parameters
        ----------
        key : str
        initial_conditions: list[dict]
        """
    # endregion

    # region State
    @abstractmethod
    def read_state(self, modelrun_name, timestep, decision_iteration=None) -> List[Dict]:
        """Read list of (name, build_year) for a given model_run, timestep,
        decision

        Parameters
        ----------
        model_run_name : str
        timestep : int
        decision_iteration : int, optional

        Returns
        -------
        list[dict]
        """

    @abstractmethod
    def write_state(self, state: List[Dict],
                    modelrun_name: str,
                    timestep: int,
                    decision_iteration=None):
        """State is a list of decisions with name and build_year.

        State is output from the DecisionManager

        Parameters
        ----------
        state : list[dict]
        model_run_name : str
        timestep : int
        decision_iteration : int, optional
        """
    # endregion

    # region Conversion coefficients
    @abstractmethod
    def read_coefficients(self, source_dim, destination_dim):
        """Reads coefficients from the store

        Coefficients are uniquely identified by their source/destination dimensions.
        This method and `write_coefficients` implement caching of conversion
        coefficients between a single pair of dimensions.

        Parameters
        ----------
        source_dim : str
            dimension name
        destination_dim : str
            dimension name

        Returns
        -------
        numpy.ndarray

        Notes
        -----
        To be called from :class:`~smif.convert.adaptor.Adaptor` implementations.
        """

    @abstractmethod
    def write_coefficients(self, source_dim, destination_dim, data):
        """Writes coefficients to the store

        Coefficients are uniquely identified by their source/destination dimensions.
        This method and `read_coefficients` implement caching of conversion
        coefficients between a single pair of dimensions.

        Parameters
        ----------
        source_dim : str
            dimension name
        destination_dim : str
            dimension name
        data : numpy.ndarray

        Notes
        -----
        To be called from :class:`~smif.convert.adaptor.Adaptor` implementations.
        """
    # endregion

    # region Results
    @abstractmethod
    def read_results(self, modelrun_name, model_name, output_spec, timestep=None,
                     decision_iteration=None) -> DataArray:
        """Return results of a model from a model_run for a given output at a timestep and
        decision iteration

        Parameters
        ----------
        model_run_id : str
        model_name : str
        output_spec : ~smif.metadata.spec.Spec
        timestep : int, default=None
        decision_iteration : int, default=None

        Returns
        -------
        ~smif.data_layer.data_array.DataArray
        """

    @abstractmethod
    def write_results(self, data, modelrun_name, model_name, timestep=None,
                      decision_iteration=None):
        """Write results of a `model_name` in `model_run_name` for a given `output_name`

        Parameters
        ----------
        data_array : ~smif.data_layer.data_array.DataArray
        model_run_id : str
        model_name : str
        timestep : int, optional
        decision_iteration : int, optional
        """

    @abstractmethod
    def delete_results(self, model_run_name, model_name, output_name, timestep=None,
                       decision_iteration=None):
        """Delete results for a single timestep/iteration of a model output in a model run

        Parameters
        ----------
        model_run_name : str
        model_name : str
        output_name : str
        timestep : int, default=None
        decision_iteration : int, default=None
        """

    @abstractmethod
    def available_results(self, modelrun_name):
        """List available results from a model run

        Returns
        -------
        list[tuple]
             Each tuple is (timestep, decision_iteration, model_name, output_name)
        """
    # endregion

    @classmethod
    def filter_on_timesteps(cls, dataframe, spec, path, timestep=None, timesteps=None):
        """Filter dataframe by timestep

        The 'timestep' dimension is treated as follows:

        If a single timestep is specified, the spec MAY include 'timestep' as a dimension. If
        so, the returned DataArray's spec will match the timestep requested. Otherwise, the
        DataArray will not include timestep as a dimension.

        If multiple timesteps are specified, the returned DataArray's spec will include a
        'timestep' dimension to match the timesteps requested.

        If timestep and timesteps are None, and the stored data has a timestep column, read all
        available timesteps. The returned DataArray's spec 'timestep' dimension will match the
        timesteps requested. If the stored data does not have a timestep column, ignore and
        pass through unchanged.
        """
        if timestep is not None:
            dataframe = cls._check_timestep_column_exists(dataframe, spec, path)
            dataframe = dataframe[dataframe.timestep == timestep]
            if 'timestep' in spec.dims:
                spec = cls._set_spec_timesteps(spec, [timestep])
            else:
                dataframe = dataframe.drop('timestep', axis=1)
        elif timesteps is not None:
            dataframe = cls._check_timestep_column_exists(dataframe, spec, path)
            dataframe = dataframe[dataframe.timestep.isin(timesteps)]
            spec = cls._set_spec_timesteps(spec, timesteps)
        elif timestep is None and timesteps is None:
            try:
                dataframe = cls._check_timestep_column_exists(dataframe, spec, path)
                spec = cls._set_spec_timesteps(spec, sorted(list(dataframe.timestep.unique())))
            except SmifDataMismatchError:
                pass

        if dataframe.empty:
            raise SmifDataNotFoundError(
                "Data for '{}' not found for timestep {}".format(spec.name, timestep))

        return dataframe, spec

    @staticmethod
    def dataframe_to_data_array(dataframe, spec, path):
        if spec.dims:
            data_array = DataArray.from_df(spec, dataframe)
        else:
            # zero-dimensional case (scalar)
            data = dataframe[spec.name]
            if data.shape != (1,):
                msg = "Data for '{}' should contain a single value, instead got {} while " + \
                        "reading from {}"
                raise SmifDataMismatchError(msg.format(spec.name, len(data), path))
            data_array = DataArray(spec, data.iloc[0])

        return data_array

    @staticmethod
    def _check_timestep_column_exists(dataframe, spec, path):
        if 'timestep' not in dataframe.columns:
            if 'timestep' in dataframe.index.names:
                dataframe = dataframe.reset_index()
            else:
                msg = "Data for '{name}' expected a column called 'timestep', instead " + \
                        "got data columns {data_columns} and index names {index_names} " + \
                        "while reading from {path}"
                raise SmifDataMismatchError(msg.format(
                    data_columns=dataframe.columns.values.tolist(),
                    index_names=dataframe.index.names,
                    name=spec.name,
                    path=path))
        return dataframe

    @staticmethod
    def _set_spec_timesteps(spec, timesteps):
        spec_config = spec.as_dict()
        if 'timestep' not in spec_config['dims']:
            spec_config['dims'] = ['timestep'] + spec_config['dims']
        spec_config['coords']['timestep'] = timesteps
        return Spec.from_dict(spec_config)



"""Results provides a common interface to access results from model runs.
"""

from typing import Union

import pandas as pd
from smif.data_layer.store import Store


class Results:
    """Common interface to access results from model runs.

    Parameters
    ----------
    store: Store or dict
        pre-created Store object or dictionary of the form {'interface': <interface>,
        'dir': <dir>} where <interface> is either 'local_csv' or 'local_parquet', and <dir> is
        the model base directory
    """

    def __init__(self, store: Union[Store, dict]):

        if type(store) is dict:
            self._store = Store.from_dict(store)
        else:
            self._store = store  # type: Store

        # keep tabs on the units of any read outputs
        self._output_units = dict()  # type: dict

    def list_model_runs(self):
        """Return a list of model run names.

        Returns
        -------
        List of model run names
        """
        return sorted([x['name'] for x in self._store.read_model_runs()])

    def list_sector_models(self, model_run_name: str):
        """Return a list of sector models for given model run.

        Parameters
        ----------
        model_run_name: str the requested model run

        Returns
        -------
        List of sector models for the given model run
        """
        return sorted(
            self._store.read_sos_model(
                self._store.read_model_run(model_run_name)['sos_model'])['sector_models']
        )

    def list_scenarios(self, model_run_name: str):
        """Return a dictionary of scenarios for given model run.

        Parameters
        ----------
        model_run_name: str the requested model run

        Returns
        -------
        Dictionary of (scenario name, variant) for the given model run.
        """
        return dict(self._store.read_model_run(model_run_name)['scenarios'])

    def list_scenario_outputs(self, scenario_name: str):
        """Return a list of outputs of a given scenario.

        Parameters
        ----------
        scenario_name: str the requested scenario

        Returns
        -------
        List of outputs for the requested scenario
        """
        return sorted(
            [x['name'] for x in self._store.read_scenario(scenario_name)['provides']])

    def list_outputs(self, sector_model_name: str):
        """Return a list of model run names.

        Parameters
        ----------
        sector_model_name: str the requested sector model

        Returns
        -------
        List of outputs for the given sector model
        """
        return sorted(
            [x['name'] for x in self._store.read_model(sector_model_name)['outputs']])

    def available_results(self, model_run_name):
        """Return the results available for a given model run.

        Parameters
        ----------
        model_run_name: str the requested model run

        Returns
        -------
        A nested dictionary data structure of the results available for the given model run
        """

        available = self._store.available_results(model_run_name)
        model_run = self._store.read_model_run(model_run_name)

        results = {
            'model_run': model_run_name,
            'sos_model': model_run['sos_model'],
            'sector_models': dict(),
            'scenarios': dict(model_run['scenarios'])
        }

        model_names = {sec for _t, _d, sec, _out in available}
        for model_name in model_names:
            results['sector_models'][model_name] = {
                'outputs': dict(),
            }

            outputs = {out for _t, _d, sec, out in available if sec == model_name}

            for output in outputs:
                results['sector_models'][model_name]['outputs'][output] = dict()

                decs = {d for _t, d, sec, out in available if
                        sec == model_name and out == output}

                for dec in decs:
                    ts = sorted({t for t, d, sec, out in available if
                                 d == dec and sec == model_name and out == output})
                    results['sector_models'][model_name]['outputs'][output][dec] = ts

        return results

    def read_results(
            self,
            model_run_names: list,
            model_names: list,
            output_names: list,
            timesteps: list = None,
            decisions: list = None,
            time_decision_tuples: list = None,
    ):
        """Return results from the store as a formatted pandas data frame. There are a number
        of ways of requesting specific timesteps/decisions. You can specify either:

            a list of (timestep, decision) tuples
                in which case data for all of those tuples matching the available results will
                be returned
        or:
            a list of timesteps
                in which case data for all of those timesteps (and any decision iterations)
                matching the available results will be returned
        or:
            a list of decision iterations
                in which case data for all of those decision iterations (and any timesteps)
                matching the available results will be returned
        or:
            a list of timesteps and a list of decision iterations
                in which case data for the Cartesian product of those timesteps and those
                decision iterations matching the available results will be returned
        or:
            nothing
                in which case all available results will be returned

        Parameters
        ----------
        model_run_names: list
            the requested model run names
        model_names: list
            the requested sector model names (exactly one required)
        output_names: list
            the requested output names (output specs must all match)
        timesteps: list
            the requested timesteps
        decisions: list
            the requested decision iterations
        time_decision_tuples: list
            a list of requested (timestep, decision) tuples

        Raises
        ------
        SmifDataNotFoundError
            If data cannot be found in the store when try to read from the store
        SmifDataMismatchError
            Data presented to read, write and update methods is in the
            incorrect format or of wrong dimensions to that expected
        SmifDataReadError
            When unable to read data e.g. unable to handle file type or connect
            to database

        Returns
        -------
        pandas.DataFrame
        """

        self.validate_names(model_run_names, model_names, output_names)

        results_dict = self._store.get_results(
            model_run_names,
            model_names[0],
            output_names,
            timesteps,
            decisions,
            time_decision_tuples
        )

        # Keep tabs on the units for each output
        for model_run_name in model_run_names:
            for output_name in output_names:
                res = results_dict[model_run_name][output_name]
                self._output_units[res.name] = res.unit

        # For each output, concatenate all requested model runs into a single data frame
        formatted_frames = []
        for output_name in output_names:
            # Get each DataArray as a pandas data frame and concatenate, resetting the index to
            # give back a flat data array
            list_of_df = [results_dict[x][output_name].as_df() for x in model_run_names]
            names_of_df = [x for x in results_dict.keys()]

            formatted_frames.append(
                pd.concat(list_of_df, keys=names_of_df, names=['model_run']).reset_index())

        # Append the other output columns to the first data frame
        formatted_frame = formatted_frames.pop(0)
        output_names.pop(0)

        for other_frame, output_name in zip(formatted_frames, output_names):
            assert (formatted_frame['model_run'] == other_frame['model_run']).all()
            assert (formatted_frame['timestep_decision'] == other_frame[
                'timestep_decision']).all()
            formatted_frame[output_name] = other_frame[output_name]

        # Unpack the timestep_decision tuples into individual columns and drop the combined
        formatted_frame[['timestep', 'decision']] = pd.DataFrame(
            formatted_frame['timestep_decision'].tolist(), index=formatted_frame.index)

        formatted_frame = formatted_frame.drop(columns=['timestep_decision'])

        # Now reorder the columns. Want model_run then timestep then decision
        cols = formatted_frame.columns.tolist()

        assert (cols[0] == 'model_run')
        cols.insert(1, cols.pop(cols.index('timestep')))
        cols.insert(2, cols.pop(cols.index('decision')))
        assert (cols[0:3] == ['model_run', 'timestep', 'decision'])

        return formatted_frame[cols]

    def read_scenario_data(self, scenario_name: str, variant_name: str, variable_name: str,
                           timesteps: list) -> pd.DataFrame:
        """Return scenario variant data from the store as a formatted pandas data frame.

        Parameters
        ----------
        scenario_name: str
            the requested scenario name
        variant_name: str
            the requested scenario variant name
        variable_name: str
            the requested output variable name that the requested scenario provides
        timesteps: list
            the requested timesteps

        Raises
        ------
        SmifDataNotFoundError
            If data cannot be found in the store when try to read from the store
        SmifDataMismatchError
            Data presented to read, write and update methods is in the
            incorrect format or of wrong dimensions to that expected
        SmifDataReadError
            When unable to read data e.g. unable to handle file type or connect
            to database

        Returns
        -------
        pandas.DataFrame
        """

        # Query the store and return as pandas data frame sorted with ascending timestep
        scenario_data_frame = self._store.read_scenario_variant_data(
            scenario_name=scenario_name,
            variant_name=variant_name,
            variable=variable_name,
            timesteps=timesteps
        ).as_df().sort_values('timestep').reset_index()

        # Reorder the columns with timestep left-most
        cols = scenario_data_frame.columns.tolist()
        assert 'timestep' in cols
        cols.insert(0, cols.pop(cols.index('timestep')))

        return scenario_data_frame[cols]

    def get_units(self, output_name: str):
        """ Return the units of a given output.

        Parameters
        ----------
        output_name: str

        Returns
        -------
        str
        """
        return self._output_units[output_name]

    def validate_names(self, model_run_names, sec_model_names, output_names):

        if len(sec_model_names) != 1:
            raise NotImplementedError(
                'Results.read() currently requires exactly one sector model'
            )

        if len(model_run_names) < 1:
            raise ValueError(
                'Results.read() requires at least one sector model name'
            )

        if len(output_names) < 1:
            raise ValueError(
                'Results.read() requires at least one output name'
            )



"""Test command line interface
"""

import os
import subprocess
import sys
from itertools import product
from tempfile import TemporaryDirectory
from time import sleep
from unittest.mock import call, patch

import smif
from pytest import fixture
from smif.cli import confirm, parse_arguments, setup_project_folder


@fixture
def tmp_sample_project(tmpdir_factory):
    test_folder = tmpdir_factory.mktemp("smif")
    subprocess.run(
        ["smif", "setup", "-d", str(test_folder), "-v"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    return str(test_folder)


def get_args(args):
    """Get args object from list of strings
    """
    parser = parse_arguments()
    return parser.parse_args(args)


def test_parse_arguments():
    """Setup a project folder argument parsing
    """
    with TemporaryDirectory() as project_folder:
        args = get_args(['setup', '-d', project_folder])

        expected = project_folder
        actual = args.directory
        assert actual == expected

        # Ensure that the `setup_project_folder` function is called when `setup`
        # command is passed to the cli
        assert args.func.__name__ == 'setup_project_folder'


def test_fixture_single_run(tmp_sample_project):
    """Test running the (default) binary-filesystem-based single_run fixture
    """
    config_dir = tmp_sample_project
    output = subprocess.run(["smif", "run", "-d", config_dir,
                             "energy_central", "-v"],
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    print(output.stdout.decode("utf-8"))
    print(output.stderr.decode("utf-8"), file=sys.stderr)
    assert "Running energy_central" in str(output.stderr)
    assert "Model run 'energy_central' complete" in str(output.stdout)


def test_fixture_single_run_csv(tmp_sample_project):
    """Test running the csv-filesystem-based single_run fixture
    """
    output = subprocess.run(
        ["smif", "run", "-i", "local_csv", "-d", tmp_sample_project,
         "energy_central", "-v"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    print(output.stdout.decode("utf-8"))
    print(output.stderr.decode("utf-8"), file=sys.stderr)
    assert "Running energy_central" in str(output.stderr)
    assert "Model run 'energy_central' complete" in str(output.stdout)


def test_fixture_single_run_warm(tmp_sample_project):
    """Test running the (default) single_run fixture with warm setting enabled
    """
    config_dir = tmp_sample_project
    output = subprocess.run(["smif", "run", "-v", "-w", "-d", config_dir,
                             "energy_central"],
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    print(output.stdout.decode("utf-8"))
    print(output.stderr.decode("utf-8"), file=sys.stderr)
    assert "Running energy_central" in str(output.stderr)
    assert "Model run 'energy_central' complete" in str(output.stdout)


def test_fixture_batch_run(tmp_sample_project):
    """Test running the multiple modelruns using the batch_run option
    """
    config_dir = tmp_sample_project
    output = subprocess.run(["smif", "run", "-v", "-b", "-d", config_dir,
                             os.path.join(config_dir, "batchfile")],
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    print(output.stdout.decode("utf-8"))
    print(output.stderr.decode("utf-8"), file=sys.stderr)
    assert "Running energy_water_cp_cr" in str(output.stderr)
    assert "Model run 'energy_water_cp_cr' complete" in str(output.stdout)
    assert "Running energy_central" in str(output.stderr)
    assert "Model run 'energy_central' complete" in str(output.stdout)


def test_fixture_list_runs(tmp_sample_project):
    """Test running the filesystem-based single_run fixture
    """
    config_dir = tmp_sample_project
    output = subprocess.run(["smif", "list", "-d", config_dir], stdout=subprocess.PIPE)
    assert "energy_water_cp_cr" in str(output.stdout)
    assert "energy_central" in str(output.stdout)

    # Run energy_central and re-check output with optional flag for completed results
    subprocess.run(["smif", "run", "energy_central", "-d", config_dir], stdout=subprocess.PIPE)
    output = subprocess.run(["smif", "list", "-c", "-d", config_dir], stdout=subprocess.PIPE)
    assert "energy_central *" in str(output.stdout)


def test_fixture_available_results(tmp_sample_project):
    """Test cli for listing available results
    """
    config_dir = tmp_sample_project
    output = subprocess.run(["smif", "available_results", "energy_central", "-d", config_dir],
                            stdout=subprocess.PIPE)

    out_str = str(output.stdout)
    assert out_str.count('model run: energy_central') == 1
    assert out_str.count('sos model: energy') == 1
    assert out_str.count('sector model:') == 1
    assert out_str.count('output:') == 2
    assert out_str.count('output: cost') == 1
    assert out_str.count('output: water_demand') == 1
    assert out_str.count('no results') == 2
    assert out_str.count('decision') == 0

    # Run energy_central and re-check output with optional flag for completed results
    subprocess.run(["smif", "run", "energy_central", "-d", config_dir], stdout=subprocess.PIPE)
    output = subprocess.run(["smif", "available_results", "energy_central", "-d", config_dir],
                            stdout=subprocess.PIPE)

    out_str = str(output.stdout)
    assert out_str.count('model run: energy_central') == 1
    assert out_str.count('sos model: energy') == 1
    assert out_str.count('sector model:') == 1
    assert out_str.count('output:') == 2
    assert out_str.count('output: cost') == 1
    assert out_str.count('output: water_demand') == 1
    assert out_str.count('no results') == 0
    assert out_str.count('decision') == 8
    assert out_str.count('decision 1') == 2
    assert out_str.count('decision 2') == 2
    assert out_str.count('decision 3') == 2
    assert out_str.count('decision 4') == 2
    assert out_str.count(': 2010') == 4
    assert out_str.count(': 2015') == 2
    assert out_str.count(': 2020') == 2


def test_fixture_missing_results(tmp_sample_project):
    """Test cli for listing missing results
    """
    config_dir = tmp_sample_project
    output = subprocess.run(["smif", "missing_results", "energy_central", "-d", config_dir],
                            stdout=subprocess.PIPE)

    out_str = str(output.stdout)
    assert out_str.count('model run: energy_central') == 1
    assert out_str.count('sos model: energy') == 1
    assert out_str.count('sector model:') == 1
    assert out_str.count('output:') == 2
    assert out_str.count('output: cost') == 1
    assert out_str.count('output: water_demand') == 1
    assert out_str.count('no missing results') == 0
    assert out_str.count('results missing for:') == 2

    # Run energy_central and re-check output with optional flag for completed results
    subprocess.run(["smif", "run", "energy_central", "-d", config_dir], stdout=subprocess.PIPE)
    output = subprocess.run(["smif", "missing_results", "energy_central", "-d", config_dir],
                            stdout=subprocess.PIPE)

    out_str = str(output.stdout)
    assert out_str.count('model run: energy_central') == 1
    assert out_str.count('sos model: energy') == 1
    assert out_str.count('sector model:') == 1
    assert out_str.count('output:') == 2
    assert out_str.count('output: cost') == 1
    assert out_str.count('output: water_demand') == 1
    assert out_str.count('no missing results') == 2
    assert out_str.count('results missing for:') == 0


def test_fixture_prepare_model_runs(tmp_sample_project):
    """Test cli for preparing model runs from template
    referencing scenario with 1 or more variants
    """
    config_dir = tmp_sample_project
    pop_variants = ['low', 'med', 'high']
    nb_variants = len(pop_variants)

    clear_model_runs(config_dir)

    subprocess.run(["smif", "prepare-run", "population", "energy_central", "-d", config_dir],
                   stdout=subprocess.PIPE)
    for suffix in pop_variants:
        filename = 'energy_central_population_' + suffix + '.yml'
        assert os.path.isfile(os.path.join(config_dir, 'config/model_runs', filename))

    variant_range = range(0, nb_variants)
    for s, e in product(variant_range, variant_range):
        clear_model_runs(config_dir)
        subprocess.run(["smif", "prepare-run", "population", "energy_central", "-s", str(s),
                        "-e", str(e), "-d", config_dir], stdout=subprocess.PIPE)
        for suffix in pop_variants[s:e + 1]:
            filename = 'energy_central_population_' + suffix + '.yml'
            assert os.path.isfile(os.path.join(config_dir, 'config/model_runs', filename))
        for suffix in pop_variants[0:s]:
            filename = 'energy_central_population_' + suffix + '.yml'
            assert not os.path.isfile(os.path.join(config_dir, 'config/model_runs', filename))
        if e < variant_range[-1]:
            for suffix in pop_variants[e + 1:]:
                filename = 'energy_central_population_' + suffix + '.yml'
                assert not os.path.isfile(
                    os.path.join(config_dir, 'config/model_runs', filename))


def clear_model_runs(config_dir):
    """ Helper function for test function
        test_fixture_prepare_model_runs
    """
    for suffix in ['low', 'med', 'high']:
        filename = 'energy_central_population_' + suffix + '.yml'
        if os.path.isfile(os.path.join(config_dir, 'config/model_runs', filename)):
            os.remove(os.path.join(config_dir, 'config/model_runs', filename))


def test_setup_project_folder():
    """Test contents of the setup project folder
    """
    with TemporaryDirectory() as project_folder:
        args = get_args(['setup', '-d', project_folder])
        setup_project_folder(args)

        assert os.path.exists(project_folder)

        folder_list = ['config', 'data', 'models', 'planning']
        for folder in folder_list:
            folder_path = os.path.join(project_folder, folder)

            assert os.path.exists(folder_path)


def test_prepare_convert(tmp_sample_project):
    project_folder = tmp_sample_project
    print(project_folder)
    # clean up
    # r=root, d=directories, f = files
    path = os.path.join(project_folder, 'data')
    for r, d, f in os.walk(path):
        for filename in f:
            if '.parquet' in filename:
                os.remove(os.path.join(r, filename))

    list_of_files = {
        'initial_conditions': [],
        'interventions': ['energy_supply', 'energy_supply_alt'],
        'narratives': [],
        'parameters': ['defaults'],
        'scenarios': ['population_density_low', 'population_density_med',
                      'population_density_high', 'population_low', 'population_med',
                      'population_high'],
        'strategies': ['build_nuke'],
        }

    subprocess.run(["smif", "prepare-convert", "energy_central", "-d", project_folder, "-i",
                    "local_csv"], stdout=subprocess.PIPE)
    # assert that correct files have been generated
    for folder in list_of_files.keys():
        for filename in list_of_files[folder]:
            path = os.path.join(project_folder, 'data', folder, filename)
            path = "{}.parquet".format(path)
            assert os.path.isfile(path)

    sleep(2)

    # Now call prepare-convert with the --noclobber option
    # all previously generated parquet files should not be modified
    subprocess.run(["smif", "prepare-convert", "energy_central", "--noclobber", "-d",
                    project_folder, "-i", "local_csv"], stdout=subprocess.PIPE)
    # assert that files have not been modified
    for folder in list_of_files.keys():
        for filename in list_of_files[folder]:
            path = os.path.join(project_folder, 'data', folder, filename)
            path = "{}.parquet".format(path)
            assert (os.path.getmtime(path) > 2)


@patch('builtins.input', return_value='y')
def test_confirm_yes(input):
    assert confirm()


@patch('builtins.input', return_value='n')
def test_confirm_no(input):
    assert not confirm()


@patch('builtins.input', return_value='')
def test_confirm_default_response(input):
    assert not confirm()


@patch('builtins.input', return_value='n')
@patch('builtins.print')
def test_confirm_default_message(mock_print, input):
    confirm()
    input.assert_has_calls([call('Confirm [n]|y: ')])


@patch('builtins.input', return_value='n')
@patch('builtins.print')
def test_confirm_custom_message(mock_print, input):
    confirm('Create directory?', True)
    input.assert_has_calls([call('Create directory? [y]|n: ')])


@patch('builtins.input', side_effect=['invalid', 'y'])
@patch('builtins.print')
def test_confirm_repeat_message(mock_print, input):
    confirm()
    input.assert_has_calls([call('Confirm [n]|y: '), call('Confirm [n]|y: ')])
    mock_print.assert_called_with('please enter y or n.')


def test_help():
    """Expect help from `smif` or `smif -h`
    """
    msg = "Command line tools for smif"
    output = subprocess.run(['smif'], stdout=subprocess.PIPE)
    assert msg in str(output.stdout)

    output = subprocess.run(['smif', '-h'], stdout=subprocess.PIPE)
    assert msg in str(output.stdout)


def test_version_display():
    """Expect version number from `smif -V`
    """
    output = subprocess.run(['smif', '-V'], stdout=subprocess.PIPE)
    assert smif.__version__ in str(output.stdout)


def test_verbose_debug():
    """Expect debug message from `smif -vv`
    """
    output = subprocess.run(['smif', 'list', '-vv'], stderr=subprocess.PIPE)
    assert 'DEBUG' in str(output.stderr)


def test_verbose_debug_alt():
    """Expect debug message from `smif --verbose --verbose`
    """
    output = subprocess.run(['smif', 'list', '--verbose', '--verbose'], stderr=subprocess.PIPE)
    assert 'DEBUG' in str(output.stderr)



"""Test SubProcessRunScheduler and SerialJobScheduler
"""
from unittest.mock import Mock, patch

import networkx
from pytest import fixture, raises
from smif.controller.job import SerialJobScheduler
from smif.controller.run import SubProcessRunScheduler
from smif.model import ModelOperation, ScenarioModel, SectorModel


class EmptySectorModel(SectorModel):
    def simulate(self, data):
        return data


class TestSubProcessRunScheduler():
    @patch('smif.controller.run.subprocess_run_scheduler.subprocess.Popen')
    def test_single_modelrun(self, mock_popen):
        my_scheduler = SubProcessRunScheduler()
        my_scheduler.add('my_model_run', {
            'directory': 'mock/dir',
            'verbosity': 0,
            'warm_start': False,
            'output_format': 'local_csv'
        })

        mock_popen.assert_called_with(
            'smif run  my_model_run -d mock/dir -i local_csv',
            shell=True,
            stderr=-2, stdout=-1
        )

    def test_status_modelrun_never_added(self):
        my_scheduler = SubProcessRunScheduler()
        status = my_scheduler.get_status('my_model_run')
        assert status['status'] == 'unstarted'

    @patch('smif.controller.run.subprocess_run_scheduler.subprocess.Popen')
    def test_status_model_started(self, mock_popen):
        attrs = {
            'poll.return_value': None,
            'communicate.return_value': (
                "this is a stdout".encode('utf-8'),
            ),
            'returncode': None
        }
        process_mock = Mock(**attrs)
        mock_popen.return_value = process_mock

        my_scheduler = SubProcessRunScheduler()
        my_scheduler.add('my_model_run', {
            'directory': 'mock/dir',
            'verbosity': 0,
            'warm_start': False,
            'output_format': 'local_csv'
        })
        my_scheduler.lock = True
        status = my_scheduler.get_status('my_model_run')
        assert status['status'] == 'running'

    @patch('smif.controller.run.subprocess_run_scheduler.subprocess.Popen')
    def test_status_model_done(self, mock_popen):
        attrs = {
            'poll.return_value': 0,
            'communicate.return_value': (
                "this is a stdout".encode('utf-8')
            )
        }
        process_mock = Mock(**attrs)
        mock_popen.return_value = process_mock

        my_scheduler = SubProcessRunScheduler()
        my_scheduler.add('my_model_run', {
            'directory': 'mock/dir',
            'verbosity': 0,
            'warm_start': False,
            'output_format': 'local_csv'
        })
        my_scheduler.lock = True
        response = my_scheduler.get_status('my_model_run')

        assert response['status'] == 'done'

    @patch('smif.controller.run.subprocess_run_scheduler.subprocess.Popen')
    def test_status_model_failed(self, mock_popen):
        attrs = {
            'poll.return_value': 1,
            'communicate.return_value': (
                "this is a stdout".encode('utf-8'),
            )
        }
        process_mock = Mock(**attrs)
        mock_popen.return_value = process_mock

        my_scheduler = SubProcessRunScheduler()
        my_scheduler.add('my_model_run', {
            'directory': 'mock/dir',
            'verbosity': 0,
            'warm_start': False,
            'output_format':
            'local_csv'
        })
        my_scheduler.lock = True
        response = my_scheduler.get_status('my_model_run')

        assert response['status'] == 'failed'

    @patch('smif.controller.run.subprocess_run_scheduler.subprocess.Popen')
    def test_status_model_stopped(self, mock_popen):
        attrs = {
            'poll.return_value': None,
            'communicate.return_value': (
                "this is a stdout".encode('utf-8'),
            )
        }
        process_mock = Mock(**attrs)
        mock_popen.return_value = process_mock

        my_scheduler = SubProcessRunScheduler()
        my_scheduler.add('my_model_run', {
            'directory': 'mock/dir',
            'verbosity': 0,
            'warm_start': False,
            'output_format':
            'local_csv'
        })
        my_scheduler.lock = True
        my_scheduler.kill('my_model_run')
        response = my_scheduler.get_status('my_model_run')

        assert response['status'] == 'stopped'


class TestSerialJobScheduler():
    @fixture
    def job_graph(self):
        G = networkx.DiGraph()
        a_model = ScenarioModel('a')

        G.add_node(
            'a',
            model=a_model,
            operation=ModelOperation.BEFORE_MODEL_RUN,
            modelrun_name='test',
            current_timestep=1,
            timesteps=[1],
            decision_iteration=0
        )
        b_model = EmptySectorModel('b')
        G.add_node(
            'b',
            model=b_model,
            operation=ModelOperation.SIMULATE,
            modelrun_name='test',
            current_timestep=1,
            timesteps=[1],
            decision_iteration=0
        )
        G.add_edge('a', 'b')
        return G

    @fixture
    def scheduler(self, empty_store):
        empty_store.write_model_run({
            'name': 'test',
            'narratives': {},
            'scenarios': {},
            'sos_model': 'test_sos_model'
        })
        empty_store.write_sos_model({
            'name': 'test_sos_model',
            'scenario_dependencies': [],
            'model_dependencies': []
        })
        scheduler = SerialJobScheduler(empty_store)
        return scheduler

    def test_add(self, job_graph, scheduler):
        job_id, err = scheduler.add(job_graph)

        print(err)
        assert err is None
        assert scheduler.get_status(job_id)['status'] == 'done'

    def test_default_status(self):
        scheduler = SerialJobScheduler()
        assert scheduler.get_status(0)['status'] == 'unstarted'

    def test_add_cyclic(self, job_graph, scheduler):
        job_graph.add_edge('b', 'a')
        job_id, err = scheduler.add(job_graph)

        assert isinstance(err, NotImplementedError)
        assert scheduler.get_status(job_id)['status'] == 'failed'

    def test_kill_fails(self, job_graph, scheduler):
        job_id, err = scheduler.add(job_graph)

        assert err is None
        with raises(NotImplementedError):
            scheduler.kill(job_id)

    def test_unknown_operation(self, job_graph, scheduler):
        model = EmptySectorModel('c')

        job_graph.add_node(
            'c',
            model=model,
            operation='unknown_operation',
            modelrun_name='test',
            current_timestep=1,
            timesteps=[1],
            decision_iteration=0
        )
        job_id, err = scheduler.add(job_graph)

        assert isinstance(err, ValueError)
        assert scheduler.get_status(job_id)['status'] == 'failed'



"""
Provides all the model run schedulers.
"""

from smif.controller.run.dafni_run_scheduler import DAFNIRunScheduler
# import classes for access like ::
#         from smif.controller.run import SubProcessRunScheduler
from smif.controller.run.subprocess_run_scheduler import SubProcessRunScheduler

# Define what should be imported as * ::
#         from smif.controller.run import *
__all__ = ['SubProcessRunScheduler', 'DAFNIRunScheduler']



"""The store provides a common data interface to smif configuration, data and metadata.

Raises
------
SmifDataNotFoundError
    If data cannot be found in the store when try to read from the store
SmifDataExistsError
    If data already exists in the store when trying to write to the store
    (use an update method instead)
SmifDataMismatchError
    Data presented to read, write and update methods is in the
    incorrect format or of wrong dimensions to that expected
SmifDataReadError
    When unable to read data e.g. unable to handle file type or connect
    to database
"""
import itertools
import logging
import os
from collections import OrderedDict, defaultdict
from copy import deepcopy
from operator import itemgetter
from os.path import splitext
from typing import Dict, List, Optional, Union

import numpy as np  # type: ignore
from smif.data_layer import DataArray
from smif.data_layer.abstract_data_store import DataStore
from smif.data_layer.abstract_metadata_store import MetadataStore
from smif.data_layer.file import (CSVDataStore, FileMetadataStore,
                                  ParquetDataStore, YamlConfigStore)
from smif.data_layer.validate import (validate_sos_model_config,
                                      validate_sos_model_format)
from smif.exception import SmifDataError, SmifDataNotFoundError
from smif.metadata.spec import Spec


class Store():
    """Common interface to data store, composed of config, metadata and data store
    implementations.

    Parameters
    ----------
    config_store: ~smif.data_layer.abstract_config_store.ConfigStore
    metadata_store: ~smif.data_layer.abstract_metadata_store.MetadataStore
    data_store: ~smif.data_layer.abstract_data_store.DataStore
    """

    def __init__(self, config_store, metadata_store: MetadataStore,
                 data_store: DataStore, model_base_folder="."):
        self.logger = logging.getLogger(__name__)
        self.config_store = config_store
        self.metadata_store = metadata_store
        self.data_store = data_store
        # base folder for any relative paths to models
        self.model_base_folder = str(model_base_folder)

    @classmethod
    def from_dict(cls, config):
        """Create Store from configuration dict
        """

        try:
            interface = config['interface']
        except KeyError:
            logging.warning('No interface provided for Results().  Assuming local_csv')
            interface = 'local_csv'

        try:
            directory = config['dir']
        except KeyError:
            logging.warning("No directory provided for Results().  Assuming '.'")
            directory = '.'

        # Check that the directory is valid
        if not os.path.isdir(directory):
            raise ValueError('Expected {} to be a valid directory'.format(directory))

        if interface == 'local_csv':
            data_store = CSVDataStore(directory)
        elif interface == 'local_parquet':
            data_store = ParquetDataStore(directory)
        else:
            raise ValueError(
                'Unsupported interface "{}". Supply local_csv or local_parquet'.format(
                    interface))

        return cls(
            config_store=YamlConfigStore(directory),
            metadata_store=FileMetadataStore(directory),
            data_store=data_store,
            model_base_folder=directory
        )

    #
    # CONFIG
    #

    # region Model runs
    def read_model_runs(self):
        """Read all system-of-system model runs

        Returns
        -------
        list[~smif.controller.modelrun.ModelRun]
        """
        return sorted(self.config_store.read_model_runs(), key=itemgetter('name'))

    def read_model_run(self, model_run_name):
        """Read a system-of-system model run

        Parameters
        ----------
        model_run_name : str

        Returns
        -------
        ~smif.controller.modelrun.ModelRun
        """
        return self.config_store.read_model_run(model_run_name)

    def write_model_run(self, model_run):
        """Write system-of-system model run

        Parameters
        ----------
        model_run : ~smif.controller.modelrun.ModelRun
        """
        self.config_store.write_model_run(model_run)

    def update_model_run(self, model_run_name, model_run):
        """Update system-of-system model run

        Parameters
        ----------
        model_run_name : str
        model_run : ~smif.controller.modelrun.ModelRun
        """
        self.config_store.update_model_run(model_run_name, model_run)

    def delete_model_run(self, model_run_name):
        """Delete a system-of-system model run

        Parameters
        ----------
        model_run_name : str
        """
        self.config_store.delete_model_run(model_run_name)

    # endregion

    # region System-of-systems models
    def read_sos_models(self):
        """Read all system-of-system models

        Returns
        -------
        list[~smif.model.sos_model.SosModel]
        """
        return sorted(self.config_store.read_sos_models(), key=itemgetter('name'))

    def read_sos_model(self, sos_model_name):
        """Read a specific system-of-system model

        Parameters
        ----------
        sos_model_name : str

        Returns
        -------
        ~smif.model.sos_model.SosModel
        """
        return self.config_store.read_sos_model(sos_model_name)

    def write_sos_model(self, sos_model):
        """Write system-of-system model

        Parameters
        ----------
        sos_model : ~smif.model.sos_model.SosModel
        """
        validate_sos_model_format(sos_model)
        self.config_store.write_sos_model(sos_model)

    def update_sos_model(self, sos_model_name, sos_model):
        """Update system-of-system model

        Parameters
        ----------
        sos_model_name : str
        sos_model : ~smif.model.sos_model.SosModel
        """
        models = self.config_store.read_models()
        scenarios = self.config_store.read_scenarios()
        validate_sos_model_config(sos_model, models, scenarios)

        self.config_store.update_sos_model(sos_model_name, sos_model)

    def delete_sos_model(self, sos_model_name):
        """Delete a system-of-system model

        Parameters
        ----------
        sos_model_name : str
        """
        self.config_store.delete_sos_model(sos_model_name)

    # endregion

    # region Models
    def read_models(self, skip_coords=False):
        """Read all models

        Returns
        -------
        list[~smif.model.model.Model]
        """
        models = sorted(self.config_store.read_models(), key=itemgetter('name'))
        if not skip_coords:
            models = [
                self._add_coords(model, ('inputs', 'outputs', 'parameters'))
                for model in models
            ]
        return models

    def read_model(self, model_name, skip_coords=False):
        """Read a model

        Parameters
        ----------
        model_name : str

        Returns
        -------
        ~smif.model.model.Model
        """
        model = self.config_store.read_model(model_name)
        if not skip_coords:
            model = self._add_coords(model, ('inputs', 'outputs', 'parameters'))
        return model

    def write_model(self, model):
        """Write a model

        Parameters
        ----------
        model : ~smif.model.model.Model
        """
        self.config_store.write_model(model)

    def update_model(self, model_name, model):
        """Update a model

        Parameters
        ----------
        model_name : str
        model : ~smif.model.model.Model
        """
        self.config_store.update_model(model_name, model)

    def delete_model(self, model_name):
        """Delete a model

        Parameters
        ----------
        model_name : str
        """
        self.config_store.delete_model(model_name)

    # endregion

    # region Scenarios
    def read_scenarios(self, skip_coords=False):
        """Read scenarios

        Returns
        -------
        list[~smif.model.ScenarioModel]
        """
        scenarios = sorted(self.config_store.read_scenarios(), key=itemgetter('name'))
        if not skip_coords:
            scenarios = [
                self._add_coords(scenario, ['provides'])
                for scenario in scenarios
            ]
        return scenarios

    def read_scenario(self, scenario_name, skip_coords=False):
        """Read a scenario

        Parameters
        ----------
        scenario_name : str

        Returns
        -------
        ~smif.model.ScenarioModel
        """
        scenario = self.config_store.read_scenario(scenario_name)
        if not skip_coords:
            scenario = self._add_coords(scenario, ['provides'])
        return scenario

    def write_scenario(self, scenario):
        """Write scenario

        Parameters
        ----------
        scenario : ~smif.model.ScenarioModel
        """
        self.config_store.write_scenario(scenario)

    def update_scenario(self, scenario_name, scenario):
        """Update scenario

        Parameters
        ----------
        scenario_name : str
        scenario : ~smif.model.ScenarioModel
        """
        self.config_store.update_scenario(scenario_name, scenario)

    def delete_scenario(self, scenario_name):
        """Delete scenario from project configuration

        Parameters
        ----------
        scenario_name : str
        """
        self.config_store.delete_scenario(scenario_name)

    def prepare_scenario(self, scenario_name, list_of_variants):
        """ Modify {scenario_name} config file to include multiple
        scenario variants.

        Parameters
        ----------
        scenario_name: str
        list_of_variants: list[int] - indices of scenario variants
        """
        scenario = self.read_scenario(scenario_name)
        # Check that template scenario file does not define more than one variant
        if not scenario['variants'] or len(scenario['variants']) > 1:
            raise SmifDataError("Template scenario file must define one"
                                " unique template variant.")

        # Read variant defined in template scenario file
        variant_template_name = scenario['variants'][0]['name']
        base_variant = self.read_scenario_variant(scenario_name, variant_template_name)
        self.delete_scenario_variant(scenario_name, variant_template_name)

        # Read template names of scenario variant data files
        output_filenames = {}  # output_name => (base, ext)
        # root is a dict. keyed on scenario outputs.
        # Entries contain the root of the variants filenames
        for output in scenario['provides']:
            output_name = output['name']
            base, ext = splitext(base_variant['data'][output_name])
            output_filenames[output_name] = base, ext
        # Now modify scenario file
        for variant_number in list_of_variants:
            # Copying the variant dict is required when underlying config_store
            # is an instance of MemoryConfigStore, which attribute _scenarios holds
            # a reference to the variant object passed to update or
            # write_scenario_variant
            variant = deepcopy(base_variant)
            variant['name'] = '{}_{:03d}'.format(scenario_name, variant_number)
            variant['description'] = '{} variant number {:03d}'.format(
                scenario_name, variant_number)
            for output_name, (base, ext) in output_filenames.items():
                variant['data'][output_name] = '{}{:03d}{}'.format(base, variant_number, ext)
            self.write_scenario_variant(scenario_name, variant)

    def prepare_model_runs(self, model_run_name, scenario_name,
                           first_var, last_var):
        """Write multiple model run config files corresponding to multiple
        scenario variants of {scenario_name}, based on template {model_run_name}
           Write batchfile containing each of the generated model runs

        Parameters
        ----------
        model_run_name: str
        scenario_name: str
        first_var: int - between 0 and number of variants-1
        last_var: int - between first_var and number of variants-1
        """

        model_run = self.read_model_run(model_run_name)
        scenario = self.read_scenario(scenario_name)

        # read strategies from config store (Store.read_strategies pulls together data on
        # interventions as well, which we don't need here)
        config_strategies = self.config_store.read_strategies(model_run_name)
        # Open batchfile
        f_handle = open(model_run_name + '.batch', 'w')
        # For each variant model_run, write a new model run file with corresponding
        # scenario variant and update batchfile
        for variant in scenario['variants'][first_var:last_var + 1]:
            variant_model_run_name = model_run_name + '_' + variant['name']
            model_run_copy = deepcopy(model_run)
            model_run_copy['name'] = variant_model_run_name
            model_run_copy['scenarios'][scenario_name] = variant['name']

            self.write_model_run(model_run_copy)
            self.config_store.write_strategies(variant_model_run_name, config_strategies)
            f_handle.write(model_run_name + '_' + variant['name'] + '\n')

        # Close batchfile
        f_handle.close()

    # endregion

    # region Scenario Variants
    def read_scenario_variants(self, scenario_name):
        """Read variants of a given scenario

        Parameters
        ----------
        scenario_name : str

        Returns
        -------
        list[dict]
        """
        scenario_variants = self.config_store.read_scenario_variants(scenario_name)
        return sorted(scenario_variants, key=itemgetter('name'))

    def read_scenario_variant(self, scenario_name, variant_name):
        """Read a scenario variant

        Parameters
        ----------
        scenario_name : str
        variant_name : str

        Returns
        -------
        dict
        """
        return self.config_store.read_scenario_variant(scenario_name, variant_name)

    def write_scenario_variant(self, scenario_name, variant):
        """Write scenario to project configuration

        Parameters
        ----------
        scenario_name : str
        variant : dict
        """
        self.config_store.write_scenario_variant(scenario_name, variant)

    def update_scenario_variant(self, scenario_name, variant_name, variant):
        """Update scenario to project configuration

        Parameters
        ----------
        scenario_name : str
        variant_name : str
        variant : dict
        """
        self.config_store.update_scenario_variant(scenario_name, variant_name, variant)

    def delete_scenario_variant(self, scenario_name, variant_name):
        """Delete scenario from project configuration

        Parameters
        ----------
        scenario_name : str
        variant_name : str
        """
        self.config_store.delete_scenario_variant(scenario_name, variant_name)

    # endregion

    # region Narratives
    def read_narrative(self, sos_model_name, narrative_name):
        """Read narrative from sos_model

        Parameters
        ----------
        sos_model_name : str
        narrative_name : str
        """
        return self.config_store.read_narrative(sos_model_name, narrative_name)

    # endregion

    # region Strategies
    def read_strategies(self, model_run_name):
        """Read strategies for a given model run

        Parameters
        ----------
        model_run_name : str

        Returns
        -------
        list[dict]
        """
        strategies = deepcopy(self.config_store.read_strategies(model_run_name))
        for strategy in strategies:
            if strategy['type'] == 'pre-specified-planning':
                strategy['interventions'] = self.data_store.read_strategy_interventions(
                    strategy)
        return strategies

    def write_strategies(self, model_run_name, strategies):
        """Write strategies for a given model_run

        Parameters
        ----------
        model_run_name : str
        strategies : list[dict]
        """
        self.config_store.write_strategies(model_run_name, strategies)

    def convert_strategies_data(self, model_run_name, tgt_store, noclobber=False):
        strategies = self.read_strategies(model_run_name)
        for strategy in strategies:
            if strategy['type'] == 'pre-specified-planning':
                data_exists = tgt_store.read_strategy_interventions(
                    strategy, assert_exists=True)
                if not(noclobber and data_exists):
                    data = self.read_strategy_interventions(strategy)
                    tgt_store.write_strategy_interventions(strategy, data)

    # endregion

    #
    # METADATA
    #

    # region Units
    def read_unit_definitions(self) -> List[str]:
        """Reads custom unit definitions

        Returns
        -------
        list[str]
            Pint-compatible unit definitions
        """
        return self.metadata_store.read_unit_definitions()

    def write_unit_definitions(self, definitions):
        """Reads custom unit definitions

        Parameters
        ----------
        list[str]
            Pint-compatible unit definitions
        """
        self.metadata_store.write_unit_definitions(definitions)

    # endregion

    # region Dimensions
    def read_dimensions(self, skip_coords=False):
        """Read dimensions

        Returns
        -------
        list[~smif.metadata.coords.Coords]
        """
        return self.metadata_store.read_dimensions(skip_coords)

    def read_dimension(self, dimension_name, skip_coords=False):
        """Return dimension

        Parameters
        ----------
        dimension_name : str

        Returns
        -------
        ~smif.metadata.coords.Coords
            A dimension definition (including elements)
        """
        return self.metadata_store.read_dimension(dimension_name, skip_coords)

    def write_dimension(self, dimension):
        """Write dimension to project configuration

        Parameters
        ----------
        dimension : ~smif.metadata.coords.Coords
        """
        self.metadata_store.write_dimension(dimension)

    def update_dimension(self, dimension_name, dimension):
        """Update dimension

        Parameters
        ----------
        dimension_name : str
        dimension : ~smif.metadata.coords.Coords
        """
        self.metadata_store.update_dimension(dimension_name, dimension)

    def delete_dimension(self, dimension_name):
        """Delete dimension

        Parameters
        ----------
        dimension_name : str
        """
        self.metadata_store.delete_dimension(dimension_name)

    def _add_coords(self, item, keys):
        """Add coordinates to spec definitions on an object
        """
        item = deepcopy(item)
        for key in keys:
            spec_list = item[key]
            for spec in spec_list:
                if 'dims' in spec and spec['dims']:
                    spec['coords'] = {
                        dim: self.read_dimension(dim)['elements']
                        for dim in spec['dims']
                    }
        return item

    # endregion

    #
    # DATA
    #

    # region Scenario Variant Data
    def read_scenario_variant_data(
            self, scenario_name: str, variant_name: str, variable: str,
            timestep: Optional[int] = None, timesteps: Optional[List[int]] = None,
            assert_exists: bool = False) -> Union[DataArray, bool]:
        """Read scenario data file

        Parameters
        ----------
        scenario_name : str
        variant_name : str
        variable : str
        timestep : int

        Returns
        -------
        data : ~smif.data_layer.data_array.DataArray
        """
        variant = self.read_scenario_variant(scenario_name, variant_name)
        key = self._key_from_data(variant['data'][variable], scenario_name, variant_name,
                                  variable)
        scenario = self.read_scenario(scenario_name)
        spec_dict = _pick_from_list(scenario['provides'], variable)
        spec = Spec.from_dict(spec_dict)
        if assert_exists:
            return self.data_store.scenario_variant_data_exists(key)
        else:
            return self.data_store.read_scenario_variant_data(key, spec, timestep, timesteps)

    def write_scenario_variant_data(self, scenario_name, variant_name, data):
        """Write scenario data file

        Parameters
        ----------
        scenario_name : str
        variant_name : str
        data : ~smif.data_layer.data_array.DataArray
        """
        variant = self.read_scenario_variant(scenario_name, variant_name)
        key = self._key_from_data(variant['data'][data.spec.name], scenario_name, variant_name,
                                  data.spec.name)
        self.data_store.write_scenario_variant_data(key, data)

    def convert_scenario_data(self, model_run_name, tgt_store, noclobber=False):
        model_run = self.read_model_run(model_run_name)
        # Convert scenario data for model run
        for scenario_name in model_run['scenarios']:
            for variant in self.read_scenario_variants(scenario_name):
                for variable in variant['data']:
                    data_exists = tgt_store.read_scenario_variant_data(
                        scenario_name, variant['name'], variable, assert_exists=True)
                    if not(noclobber and data_exists):
                        data_array = self.read_scenario_variant_data(
                            scenario_name, variant['name'], variable)
                        tgt_store.write_scenario_variant_data(
                            scenario_name, variant['name'], data_array)
    # endregion

    # region Narrative Data
    def read_narrative_variant_data(self, sos_model_name, narrative_name, variant_name,
                                    parameter_name, timestep=None, assert_exists=False):
        """Read narrative data file

        Parameters
        ----------
        sos_model_name : str
        narrative_name : str
        variant_name : str
        parameter_name : str
        timestep : int (optional)
            If None, read data for all timesteps

        Returns
        -------
        ~smif.data_layer.data_array.DataArray
        """
        sos_model = self.read_sos_model(sos_model_name)

        narrative = _pick_from_list(sos_model['narratives'], narrative_name)

        if narrative is None:
            msg = "Narrative name '{}' does not exist in sos_model '{}'"
            raise SmifDataNotFoundError(msg.format(narrative_name, sos_model_name))

        variant = _pick_from_list(narrative['variants'], variant_name)

        if variant is None:
            msg = "Variant name '{}' does not exist in narrative '{}'"
            raise SmifDataNotFoundError(msg.format(variant_name, narrative_name))

        key = self._key_from_data(variant['data'][parameter_name], narrative_name,
                                  variant_name, parameter_name)

        if assert_exists:
            return self.data_store.narrative_variant_data_exists(key)
        else:
            spec_dict = None
            # find sector model which needs this parameter, to get spec definition
            for model_name, params in narrative['provides'].items():
                if parameter_name in params:
                    sector_model = self.read_model(model_name)
                    spec_dict = _pick_from_list(sector_model['parameters'], parameter_name)
                    break
            # find spec
            if spec_dict is None:
                raise SmifDataNotFoundError("Parameter {} not found in any of {}".format(
                    parameter_name, sos_model['sector_models']))
            spec = Spec.from_dict(spec_dict)

            return self.data_store.read_narrative_variant_data(key, spec, timestep)

    def write_narrative_variant_data(self, sos_model_name, narrative_name, variant_name, data):
        """Read narrative data file

        Parameters
        ----------
        sos_model_name : str
        narrative_name : str
        variant_name : str
        data : ~smif.data_layer.data_array.DataArray
        """
        sos_model = self.read_sos_model(sos_model_name)
        narrative = _pick_from_list(sos_model['narratives'], narrative_name)
        variant = _pick_from_list(narrative['variants'], variant_name)
        key = self._key_from_data(
            variant['data'][data.spec.name], narrative_name, variant_name, data.spec.name)
        self.data_store.write_narrative_variant_data(key, data)

    def convert_narrative_data(self, sos_model_name, tgt_store, noclobber=False):
        sos_model = self.read_sos_model(sos_model_name)
        for narrative in sos_model['narratives']:
            for variant in narrative['variants']:
                for param in variant['data']:
                    data_exists = tgt_store.read_narrative_variant_data(sos_model_name,
                                                                        narrative['name'],
                                                                        variant['name'],
                                                                        param,
                                                                        assert_exists=True)
                if not(noclobber and data_exists):
                    data_array = self.read_narrative_variant_data(sos_model_name,
                                                                  narrative['name'],
                                                                  variant['name'],
                                                                  param)
                    tgt_store.write_narrative_variant_data(sos_model_name, narrative['name'],
                                                           variant['name'], data_array)

    def read_model_parameter_default(self, model_name, parameter_name, assert_exists=False):
        """Read default data for a sector model parameter

        Parameters
        ----------
        model_name : str
        parameter_name : str

        Returns
        -------
        ~smif.data_layer.data_array.DataArray
        """
        model = self.read_model(model_name)
        param = _pick_from_list(model['parameters'], parameter_name)
        spec = Spec.from_dict(param)
        try:
            path = param['default']
        except TypeError:
            raise SmifDataNotFoundError("Parameter {} not found in model {}".format(
                parameter_name, model_name))
        except KeyError:
            path = 'default__{}__{}.csv'.format(model_name, parameter_name)
        key = self._key_from_data(path, model_name, parameter_name)
        if assert_exists:
            return self.data_store.model_parameter_default_data_exists(key)
        else:
            return self.data_store.read_model_parameter_default(key, spec)

    def write_model_parameter_default(self, model_name, parameter_name, data):
        """Write default data for a sector model parameter

        Parameters
        ----------
        model_name : str
        parameter_name : str
        data : ~smif.data_layer.data_array.DataArray
        """
        model = self.read_model(model_name, skip_coords=True)
        param = _pick_from_list(model['parameters'], parameter_name)
        try:
            path = param['default']
        except TypeError:
            raise SmifDataNotFoundError("Parameter {} not found in model {}".format(
                parameter_name, model_name))
        except KeyError:
            path = 'default__{}__{}.csv'.format(model_name, parameter_name)
        key = self._key_from_data(path, model_name, parameter_name)
        self.data_store.write_model_parameter_default(key, data)

    def convert_model_parameter_default_data(self, sector_model_name, tgt_store,
                                             noclobber=False):
        sector_model = self.read_model(sector_model_name)
        for parameter in sector_model['parameters']:
            data_exists = tgt_store.read_model_parameter_default(sector_model_name,
                                                                 parameter['name'],
                                                                 assert_exists=True)
            if not(noclobber and data_exists):
                data_array = self.read_model_parameter_default(sector_model_name,
                                                               parameter['name'])
                tgt_store.write_model_parameter_default(sector_model_name, parameter['name'],
                                                        data_array)
    # endregion

    # region Interventions
    def read_interventions(self, model_name):
        """Read interventions data for `model_name`

        Returns
        -------
        dict[str, dict]
            A dict of intervention dictionaries containing intervention
            attributes keyed by intervention name
        """
        model = self.read_model(model_name, skip_coords=True)
        if model['interventions'] != []:
            return self.data_store.read_interventions(model['interventions'])
        else:
            return {}

    def write_interventions(self, model_name, interventions):
        """Write interventions data for a model

        Parameters
        ----------
        dict[str, dict]
            A dict of intervention dictionaries containing intervention
            attributes keyed by intervention name
        """
        model = self.read_model(model_name)
        model['interventions'] = [model_name + '.csv']
        self.update_model(model_name, model)
        self.data_store.write_interventions(model['interventions'][0], interventions)

    def write_interventions_file(self, model_name, string_id, interventions):
        model = self.read_model(model_name)
        if string_id in model['interventions']:
            self.data_store.write_interventions(string_id, interventions)
        else:
            raise SmifDataNotFoundError("Intervention {} not found for"
                                        " sector model {}.".format(string_id, model_name))

    def read_interventions_file(self, model_name, string_id, assert_exists=False):
        model = self.read_model(model_name)
        if string_id in model['interventions']:
            if assert_exists:
                return self.data_store.interventions_data_exists(string_id)
            else:
                return self.data_store.read_interventions([string_id])
        else:
            raise SmifDataNotFoundError("Intervention {} not found for"
                                        " sector model {}.".format(string_id, model_name))

    def convert_interventions_data(self, sector_model_name, tgt_store, noclobber=False):
        sector_model = self.read_model(sector_model_name)
        for intervention in sector_model['interventions']:
            data_exists = tgt_store.read_interventions_file(sector_model_name,
                                                            intervention,
                                                            assert_exists=True)
            if not(noclobber and data_exists):
                interventions = self.read_interventions_file(sector_model_name, intervention)
                tgt_store.write_interventions_file(
                    sector_model_name, intervention, interventions)

    def read_strategy_interventions(self, strategy, assert_exists=False):
        """Read interventions as defined in a model run strategy
        """
        if assert_exists:
            return self.data_store.strategy_data_exists(strategy)
        else:
            return self.data_store.read_strategy_interventions(strategy)

    def write_strategy_interventions(self, strategy, data):
        """
        Parameters
        ----------
        list[dicts]
        """
        self.data_store.write_strategy_interventions(strategy, data)

    def read_initial_conditions(self, model_name) -> List[Dict]:
        """Read historical interventions for `model_name`

        Returns
        -------
        list[dict]
            A list of historical interventions, with keys 'name' and 'build_year'
        """
        model = self.read_model(model_name)
        if model['initial_conditions'] != []:
            return self.data_store.read_initial_conditions(model['initial_conditions'])
        else:
            return []

    def write_initial_conditions(self, model_name, initial_conditions):
        """Write historical interventions for a model

        Parameters
        ----------
        list[dict]
            A list of historical interventions, with keys 'name' and 'build_year'
        """
        model = self.read_model(model_name)
        model['initial_conditions'] = [model_name + '.csv']
        self.update_model(model_name, model)
        self.data_store.write_initial_conditions(model['initial_conditions'][0],
                                                 initial_conditions)

    def write_initial_conditions_file(self, model_name, string_id, initial_conditions):
        model = self.read_model(model_name)
        if string_id in model['initial_conditions']:
            self.data_store.write_initial_conditions(string_id, initial_conditions)
        else:
            raise SmifDataNotFoundError("Initial condition {} not found for"
                                        " sector model {}.".format(string_id, model_name))

    def read_initial_conditions_file(self, model_name, string_id, assert_exists=False):
        model = self.read_model(model_name)
        if string_id in model['initial_conditions']:
            if assert_exists:
                return self.data_store.initial_conditions_data_exists(string_id)
            else:
                return self.data_store.read_initial_conditions([string_id])
        else:
            raise SmifDataNotFoundError("Initial conditions {} not found for"
                                        " sector model {}.".format(string_id, model_name))

    def convert_initial_conditions_data(self, sector_model_name, tgt_store, noclobber=False):
        sector_model = self.read_model(sector_model_name)
        for initial_condition in sector_model['initial_conditions']:
            data_exists = tgt_store.read_initial_conditions_file(sector_model_name,
                                                                 initial_condition,
                                                                 assert_exists=True)
            if not(noclobber and data_exists):
                initial_conditions = self.read_initial_conditions_file(sector_model_name,
                                                                       initial_condition)
                tgt_store.write_initial_conditions_file(sector_model_name, initial_condition,
                                                        initial_conditions)

    def read_all_initial_conditions(self, model_run_name) -> List[Dict]:
        """A list of all historical interventions

        Returns
        -------
        list[dict]
        """
        historical_interventions = []  # type: List
        model_run = self.read_model_run(model_run_name)
        sos_model_name = model_run['sos_model']
        sos_model = self.read_sos_model(sos_model_name)
        sector_model_names = sos_model['sector_models']
        for sector_model_name in sector_model_names:
            historical_interventions.extend(
                self.read_initial_conditions(sector_model_name)
            )
        return historical_interventions

    # endregion

    # region State
    def read_state(self, model_run_name, timestep, decision_iteration=None) -> List[Dict]:
        """Read list of (name, build_year) for a given model_run, timestep,
        decision

        Parameters
        ----------
        model_run_name : str
        timestep : int
        decision_iteration : int, optional

        Returns
        -------
        list[dict]
        """
        return self.data_store.read_state(model_run_name, timestep, decision_iteration)

    def write_state(self, state, model_run_name, timestep, decision_iteration=None):
        """State is a list of decisions with name and build_year.

        State is output from the DecisionManager

        Parameters
        ----------
        state : list[dict]
        model_run_name : str
        timestep : int
        decision_iteration : int, optional
        """
        self.data_store.write_state(state, model_run_name, timestep, decision_iteration)

    # endregion

    # region Conversion coefficients
    def read_coefficients(self, source_dim: str, destination_dim: str) -> np.ndarray:
        """Reads coefficients from the store

        Coefficients are uniquely identified by their source/destination dimensions.
        This method and `write_coefficients` implement caching of conversion
        coefficients between dimensions.

        Parameters
        ----------
        source_dim : str
            Dimension name
        destination_dim : str
            Dimension name

        Returns
        -------
        numpy.ndarray

        Notes
        -----
        To be called from :class:`~smif.convert.adaptor.Adaptor` implementations.
        """
        return self.data_store.read_coefficients(source_dim, destination_dim)

    def write_coefficients(self, source_dim: str, destination_dim: str, data: np.ndarray):
        """Writes coefficients to the store

        Coefficients are uniquely identified by their source/destination dimensions.
        This method and `read_coefficients` implement caching of conversion
        coefficients between dimensions.

        Parameters
        ----------
        source_dim : str
            Dimension name
        destination_dim : str
            Dimension name
        data : numpy.ndarray

        Notes
        -----
        To be called from :class:`~smif.convert.adaptor.Adaptor` implementations.
        """
        self.data_store.write_coefficients(source_dim, destination_dim, data)

    # endregion

    # region Results
    def read_results(self,
                     model_run_name: str,
                     model_name: str,
                     output_spec: Spec,
                     timestep: Optional[int] = None,
                     decision_iteration: Optional[int] = None) -> DataArray:
        """Return results of a `model_name` in `model_run_name` for a given `output_name`

        Parameters
        ----------
        model_run_name : str
        model_name : str
        output_spec : smif.metadata.Spec
        timestep : int, default=None
        decision_iteration : int, default=None

        Returns
        -------
        ~smif.data_layer.data_array.DataArray
        """
        return self.data_store.read_results(
            model_run_name, model_name, output_spec, timestep, decision_iteration)

    def write_results(self, data_array, model_run_name, model_name, timestep=None,
                      decision_iteration=None):
        """Write results of a `model_name` in `model_run_name` for a given `output_name`

        Parameters
        ----------
        data_array : ~smif.data_layer.data_array.DataArray
        model_run_id : str
        model_name : str
        timestep : int, optional
        decision_iteration : int, optional
        """
        self.data_store.write_results(
            data_array, model_run_name, model_name, timestep, decision_iteration)

    def delete_results(self, model_run_name, model_name, output_name, timestep=None,
                       decision_iteration=None):
        """Delete results for a single timestep/iteration of a model output in a model run

        Parameters
        ----------
        model_run_name : str
        model_name : str
        output_name : str
        timestep : int, default=None
        decision_iteration : int, default=None
        """
        self.data_store.delete_results(
            model_run_name, model_name, output_name, timestep, decision_iteration)

    def clear_results(self, model_run_name):
        """Clear all results from a single model run

        Parameters
        ----------
        model_run_name : str
        """
        available = self.available_results(model_run_name)
        for timestep, decision_iteration, model_name, output_name in available:
            self.data_store.delete_results(
                model_run_name, model_name, output_name, timestep, decision_iteration)

    def available_results(self, model_run_name):
        """List available results from a model run

        Parameters
        ----------
        model_run_name : str

        Returns
        -------
        list[tuple]
             Each tuple is (timestep, decision_iteration, model_name, output_name)
        """
        return self.data_store.available_results(model_run_name)

    def completed_jobs(self, model_run_name):
        """List completed jobs from a model run

        Parameters
        ----------
        model_run_name : str

        Returns
        -------
        list[tuple]
             Each tuple is (timestep, decision_iteration, model_name)
        """
        available_results = self.available_results(model_run_name)  # {(t, d, model, output)}
        model_outputs = self.expected_model_outputs(model_run_name)  # [(model, output)]
        completed_jobs = self.filter_complete_available_results(
            available_results, model_outputs)
        return completed_jobs

    @staticmethod
    def filter_complete_available_results(available_results, expected_model_outputs):
        """Filter available results from a model run to include only complete timestep/decision
        iteration combinations

        Parameters
        ----------
        available_results: list[tuple]
            List of (timestep, decision_iteration, model_name, output_name)
        expected_model_outputs: list[tuple]
            List or set of (model_name, output_name)

        Returns
        -------
        list[tuple]
             Each tuple is (timestep, decision_iteration, model_name)
        """
        expected_model_outputs = set(expected_model_outputs)
        model_names = {model_name for model_name, _ in expected_model_outputs}
        model_outputs_by_td = defaultdict(set)
        for timestep, decision, model_name, output_name in available_results:
            model_outputs_by_td[(timestep, decision)].add((model_name, output_name))

        completed_jobs = []
        for (timestep, decision), td_model_outputs in model_outputs_by_td.items():
            if td_model_outputs == expected_model_outputs:
                for model_name in model_names:
                    completed_jobs.append((timestep, decision, model_name))
        return completed_jobs

    def expected_model_outputs(self, model_run_name):
        """List expected model outputs from a model run

        Parameters
        ----------
        model_run_name : str

        Returns
        -------
        list[tuple]
             Each tuple is (model_name, output_name)
        """
        model_run = self.read_model_run(model_run_name)
        sos_model_name = model_run['sos_model']
        sos_config = self.read_sos_model(sos_model_name)

        # For each model, get the outputs and create (model_name, output_name) tuples
        expected_model_outputs = []
        for model_name in sos_config['sector_models']:
            model_config = self.read_model(model_name)
            for output in model_config['outputs']:
                expected_model_outputs.append((model_name, output['name']))

        return expected_model_outputs

    def prepare_warm_start(self, model_run_name):
        """Copy the results from the previous model_run if available

        The method allows a previous unsuccessful model_run to 'warm start' a new model run
        from a later timestep. Model results are recovered from the timestep that the previous
        model_run was run until, and the new model run runs from the returned timestep

        Parameters
        ----------
        model_run_name : str

        Returns
        -------
        int The timestep to which the data store was recovered

        Notes
        -----
        Called from smif.controller.execute
        """
        available_results = self.available_results(model_run_name)
        if available_results:
            max_timestep = max(
                timestep for
                timestep, decision_iteration, model_name, output_name in available_results
            )
            # could explicitly clear results for max timestep
        else:
            max_timestep = None
        return max_timestep

    def canonical_available_results(self, model_run_name):
        """List the results that are available from a model run, collapsing all decision
        iterations.

        This is the unique items from calling `available_results`, with all decision iterations
        set to 0.

        This method is used to determine whether a model run is complete, given that it is
        impossible to know how many decision iterations to expect: we simply check that each
        expected timestep has been completed.

        Parameters
        ----------
        model_run_name : str

        Returns
        -------
        set Set of tuples representing available results
        """

        available_results = self.available_results(model_run_name)

        canonical_list = []

        for t, d, model_name, output_name in available_results:
            canonical_list.append((t, 0, model_name, output_name))

        # Return as a set to remove duplicates
        return set(canonical_list)

    def canonical_expected_results(self, model_run_name):
        """List the results that are expected from a model run, collapsing all decision
        iterations.

        For a complete model run, this would coincide with the unique list returned from
        `available_results`, where all decision iterations are set to 0.

        This method is used to determine whether a model run is complete, given that it is
        impossible to know how many decision iterations to expect: we simply check that each
        expected timestep has been completed.

        Parameters
        ----------
        model_run_name : str

        Returns
        -------
        set Set of tuples representing expected results
        """

        # Model results are returned as a tuple
        # (timestep, decision_it, model_name, output_name)
        # so we first build the full list of expected results tuples.

        expected_results = []

        # Get the sos model name given the model run name, and the full list of timesteps
        model_run = self.read_model_run(model_run_name)
        timesteps = sorted(model_run['timesteps'])
        sos_model_name = model_run['sos_model']

        # Get the list of sector models in the sos model
        sos_config = self.read_sos_model(sos_model_name)

        # For each sector model, get the outputs and create the tuples
        for model_name in sos_config['sector_models']:

            model_config = self.read_model(model_name)
            outputs = model_config['outputs']

            for output, t in itertools.product(outputs, timesteps):
                expected_results.append((t, 0, model_name, output['name']))

        # Return as a set to remove duplicates
        return set(expected_results)

    def canonical_missing_results(self, model_run_name):
        """List the results that are missing from a model run, collapsing all decision
        iterations.

        For a complete model run, this is what is left after removing
        canonical_available_results from canonical_expected_results.

        Parameters
        ----------
        model_run_name : str

        Returns
        -------
        set Set of tuples representing missing results
        """

        return self.canonical_expected_results(
            model_run_name) - self.canonical_available_results(model_run_name)

    def _get_result_darray_internal(self, model_run_name, model_name, output_name,
                                    time_decision_tuples):
        """Internal implementation for `get_result_darray`, after the unique list of
        (timestep, decision) tuples has been generated and validated.

        This method gets the spec for the output defined by the model_run_name, model_name
        and output_name and expands the spec to include an additional dimension for the list of
        tuples.

        Then, for each tuple, the data array from the corresponding read_results call is
        stacked, and together with the new spec this information is returned as a new
        DataArray.

        Parameters
        ----------
        model_run_name : str
        model_name : str
        output_name : str
        time_decision_tuples : list of unique (timestep, decision) tuples

        Returns
        -------
        DataArray with expanded spec and data for each (timestep, decision) tuple
        """

        # Get the output spec given the name of the sector model and output
        output_spec = None
        model = self.read_model(model_name)

        for output in model['outputs']:

            # Ignore if the output name doesn't match
            if output_name != output['name']:
                continue

            output_spec = Spec.from_dict(output)

        assert output_spec, "Output name was not found in model outputs"

        # Read the results for each (timestep, decision) tuple and stack them
        list_of_numpy_arrays = []

        for t, d in time_decision_tuples:
            d_array = self.read_results(model_run_name, model_name, output_spec, t, d)
            list_of_numpy_arrays.append(d_array.data)

        stacked_data = np.vstack(list_of_numpy_arrays)

        # Add new dimensions to the data spec
        output_dict = output_spec.as_dict()
        output_dict['dims'] = ['timestep_decision'] + output_dict['dims']
        output_dict['coords']['timestep_decision'] = time_decision_tuples

        output_spec = Spec.from_dict(output_dict)

        # Create a new DataArray from the modified spec and stacked data
        return DataArray(output_spec, np.reshape(stacked_data, output_spec.shape))

    def get_result_darray(self, model_run_name, model_name, output_name, timesteps=None,
                          decision_iterations=None, time_decision_tuples=None):
        """Return data for multiple timesteps and decision iterations for a given output from
        a given sector model in a specific model run.

        You can specify either:
            a list of (timestep, decision) tuples
                in which case data for all of those tuples matching the available results will
                be returned
        or:
            a list of timesteps
                in which case data for all of those timesteps (and any decision iterations)
                matching the available results will be returned
        or:
            a list of decision iterations
                in which case data for all of those decision iterations (and any timesteps)
                matching the available results will be returned
        or:
            a list of timesteps and a list of decision iterations
                in which case data for the Cartesian product of those timesteps and those
                decision iterations matching the available results will be returned
        or:
            nothing
                in which case all available results will be returned

        Then, for each tuple, the data array from the corresponding read_results call is
        stacked, and together with the new spec this information is returned as a new
        DataArray.

        Parameters
        ----------
        model_run_name : str
        model_name : str
        output_name : str
        timesteps : optional list of timesteps
        decision_iterations : optional list of decision iterations
        time_decision_tuples : optional list of unique (timestep, decision) tuples

        Returns
        -------
        DataArray with expanded spec and the data requested
        """
        available = self.available_results(model_run_name)

        # Build up the necessary list of tuples
        if not timesteps and not decision_iterations and not time_decision_tuples:
            list_of_tuples = [
                (t, d) for t, d, m, out in available
                if m == model_name and out == output_name
            ]

        elif timesteps and not decision_iterations and not time_decision_tuples:
            list_of_tuples = [
                (t, d) for t, d, m, out in available
                if m == model_name and out == output_name and t in timesteps
            ]

        elif decision_iterations and not timesteps and not time_decision_tuples:
            list_of_tuples = [
                (t, d) for t, d, m, out in available
                if m == model_name and out == output_name and d in decision_iterations
            ]

        elif time_decision_tuples and not timesteps and not decision_iterations:
            list_of_tuples = [
                (t, d) for t, d, m, out in available
                if m == model_name and out == output_name and (t, d) in time_decision_tuples
            ]

        elif timesteps and decision_iterations and not time_decision_tuples:
            t_d = list(itertools.product(timesteps, decision_iterations))
            list_of_tuples = [
                (t, d) for t, d, m, out in available
                if m == model_name and out == output_name and (t, d) in t_d
            ]

        else:
            msg = "Expected either timesteps, or decisions, or (timestep, decision) " + \
                  "tuples, or timesteps and decisions, or none of the above."
            raise ValueError(msg)

        if not list_of_tuples:
            raise SmifDataNotFoundError("None of the requested data is available.")

        return self._get_result_darray_internal(
            model_run_name, model_name, output_name, sorted(list_of_tuples)
        )

    def get_results(self,
                    model_run_names: list,
                    model_name: str,
                    output_names: list,
                    timesteps: list = None,
                    decisions: list = None,
                    time_decision_tuples: list = None,
                    ):
        """Return data for multiple timesteps and decision iterations for a given output from
        a given sector model for multiple model runs.

        Parameters
        ----------
        model_run_names: list[str]
            the requested model run names
        model_name: str
            the requested sector model name
        output_names: list[str]
            the requested output names (output specs must all match)
        timesteps: list[int]
            the requested timesteps
        decisions: list[int]
            the requested decision iterations
        time_decision_tuples: list[tuple]
            a list of requested (timestep, decision) tuples

        Returns
        -------
        dict
            Nested dictionary of DataArray objects, keyed on model run name and output name.
            Returned DataArrays include one extra (timestep, decision_iteration) dimension.
        """

        # List the available output names and verify requested outputs match
        outputs = self.read_model(model_name)['outputs']
        available_outputs = [output['name'] for output in outputs]

        for output_name in output_names:
            assert output_name in available_outputs, \
                '{} is not an output of sector model {}.'.format(output_name, model_name)

        # The spec for each requested output must be the same. We check they have the same
        # coordinates
        coords = [Spec.from_dict(output).coords for output in outputs if
                  output['name'] in output_names]

        for coord in coords:
            if coord != coords[0]:
                raise ValueError('Different outputs must have the same coordinates')

        # Now actually obtain the requested results
        results_dict = OrderedDict()  # type: OrderedDict
        for model_run_name in model_run_names:
            results_dict[model_run_name] = OrderedDict()
            for output_name in output_names:
                results_dict[model_run_name][output_name] = self.get_result_darray(
                    model_run_name,
                    model_name,
                    output_name,
                    timesteps,
                    decisions,
                    time_decision_tuples
                )
        return results_dict

    # endregion

    # region data store utilities
    def _key_from_data(self, path, *args):
        """Return path or generate a unique key for a given set of args
        """
        if isinstance(self.data_store, (CSVDataStore, ParquetDataStore)):
            return path
        else:
            return tuple(args)
    # endregion


def _pick_from_list(list_of_dicts, name):
    for item in list_of_dicts:
        if 'name' in item and item['name'] == name:
            return item
    return None
